"""
This module is adapted from https://github.com/radiocosmology/caput/blob/master/caput/mpiutil.py
"""
import numpy as np
import logging

rank = 0
size = 1
_comm = None
world = None
rank0 = True


logger = logging.getLogger(__name__)

from mpi4py import MPI
_comm = MPI.COMM_WORLD
world = _comm

rank = _comm.Get_rank()
size = _comm.Get_size()

if _comm is not None and size > 1:
    logger.debug("Starting MPI rank=%i [size=%i]", rank, size)

rank0 = rank == 0

def partition_list(full_list, i, n, method="con"):
    """
    Partition a list into `n` pieces. Return the `i`th partition.
    """

    def _partition(N, n, i):
        # If partiion `N` numbers into `n` pieces,
        # return the start and stop of the `i` th piece
        base = N // n
        rem = N % n
        num_lst = rem * [base + 1] + (n - rem) * [base]
        cum_num_lst = np.cumsum([0] + num_lst)

        return cum_num_lst[i], cum_num_lst[i + 1]

    N = len(full_list)
    start, stop = _partition(N, n, i)

    if method == "con":
        return full_list[start:stop]
    elif method == "alt":
        return full_list[i::n]
    elif method == "rand":
        choices = np.random.permutation(N)[start:stop]
        return [full_list[i] for i in choices]
    else:
        raise ValueError("Unknown partition method %s" % method)

def partition_list_mpi(full_list, method="con", comm=_comm):
    """
    Return the partition of a list specific to the current MPI process.
    """
    if comm is not None:
        rank = comm.rank
        size = comm.size

    return partition_list(full_list, rank, size, method=method)

def parallel_map(func, glist, root=None, method="con", comm=_comm):
    """
    Apply a parallel map using MPI.
    Should be called collectively on the same list. All ranks return the full
    set of results.
    Parameters
    ----------
    func : function
        Function to apply.
    glist : list
        List of map over. Must be globally defined.
    root : None or Integer
        Which process should gather the results, all processes will gather the results if None.
    method: str
        How to split `glist` to each process, can be 'con': continuously, 'alt': alternatively, 'rand': randomly. Default is 'con'.
    comm : MPI communicator
        MPI communicator that array is distributed over. Default is the gobal _comm.
    Returns
    -------
    results : list
        Global list of results.
    """

    # Synchronize
    barrier(comm=comm)

    # If we're only on a single node, then just perform without MPI
    if comm is None or comm.size == 1:
        return [func(item) for item in glist]

    # Pair up each list item with its position.
    zlist = list(enumerate(glist))

    # Partition list based on MPI rank
    llist = partition_list_mpi(zlist, method=method, comm=comm)

    # Operate on sublist
    flist = [(ind, func(item)) for ind, item in llist]

    barrier(comm=comm)

    rlist = None
    if root is None:
        # Gather all results onto all ranks
        rlist = comm.allgather(flist)
    else:
        # Gather all results onto the specified rank
        rlist = comm.gather(flist, root=root)

    if rlist is not None:
        # Flatten the list of results
        flatlist = [item for sublist in rlist for item in sublist]

        # Sort into original order
        sortlist = sorted(flatlist, key=(lambda item: item[0]))

        # Synchronize
        # barrier(comm=comm)

        # Extract the return values into a list
        return [item for ind, item in sortlist]
    else:
        return None

def parallel_jobs_no_gather_no_return(func, glist, method="con", comm=_comm):
    """
    Apply a parallel map using MPI.
    Should be called collectively on the same list. All ranks return the full
    set of results.
    Parameters
    ----------
    func : function
        Function to apply.
    glist : zipped list
        List of map over. Must be globally defined.
    root : None or Integer
        Which process should gather the results, all processes will gather the results if None.
    method: str
        How to split `glist` to each process, can be 'con': continuously, 'alt': alternatively, 'rand': randomly. Default is 'con'.
    comm : MPI communicator
        MPI communicator that array is distributed over. Default is the gobal _comm.
    Returns
    -------
    results : list
        Global list of results.
    """

    # Synchronize
    barrier(comm=comm)

    # If we're only on a single node, then just perform without MPI
    if comm is None or comm.size == 1:
        return [func(item) for item in glist]


    # Partition list based on MPI rank
    llist = partition_list_mpi(glist, method=method, comm=comm)

    # Operate on sublist
    for zipped_item in llist:
        func(zip(*zipped_item))

    # Synchronize
    barrier(comm=comm)
    return None

def barrier(comm=_comm):
    """
    Synchronize all MPI processes.
    """
    if comm is not None and comm.size > 1:
        comm.Barrier()


#def save_to_hdf5_parallel(file_path, local_data, local_key):

    # Open the HDF5 file in parallel
#    with h5py.File(file_path, 'a', driver='mpio', comm=_comm) as file:
        # Create a dataset with collective mode (MPI)
#        dataset = file.create_dataset(local_key, data=local_data, chunks=True)

    # Ensure all processes have finished writing before proceeding
#    barrier()


# This is the python version of the charge_dist.pro file of spdust2.0.
# 
# It first computes the photoemission rate according to Weingartner & Draine,
# 2001b, ApJ 134, 263. All equation numbers refer to that paper.
# Then compute collisional charging rates according to Draine & Sutin, 1987, ApJ, 320, 803.                              
# Finally, gets the grain charge distribution function for given grain size      
# and environmental conditions.  


import numpy as np
from .. import SpDust_data_dir  
from ..util import cgsconst, makelogtab, DX_over_X, biinterp_func, coord_grid, log_biinterplog
from .grain_properties import asurf, N_C
from scipy.interpolate import interp1d    
import os      


class paramjpe:
    # Stores the parameters used in the JPE calculation
    W = 4.4  # Work function in eV


class Qabstabs:
    # Stores the Qabs arrays
    Nrad = None
    Nwav = None
    a_tab = None
    Qabs_hnu_tab = None
    Q_abs_ion = None
    Q_abs_neu = None

class refr_indices:
    hnu_tab = None
    la_tab = None

class jpe_arrays:
    # Stores the Jpeisrf arrays
    a_values = None
    Jpe_pos_isrf = None
    Jpe_neg_isrf = None

# Constants
c, q, k, mp, me, h, debye, eV = cgsconst.c, cgsconst.q, cgsconst.k, cgsconst.mp, cgsconst.me, cgsconst.h, cgsconst.debye, cgsconst.eV
W = paramjpe.W
pi = cgsconst.pi

# Equation (2), in eV
#@njit
def IPv(Z, a):
    return W + ((Z + 0.5) * q**2 / a + (Z + 2) * q**2 / a * (0.3e-8) / a) / eV


# Equation (7), in eV
#@njit
def E_min(Z, a):
    if Z < -1:
        return -(Z + 1) * q**2 / a / (1 + (27e-8 / a)**0.75) / eV
    else:
        return 0.0

# Equation (6), in eV
#@njit
def hnu_pet(Z, a):
    if Z >= -1:
        return max(0.0, IPv(Z, a))
    else:
        return max(0.0, IPv(Z, a) + E_min(Z, a))
    
# Equation (9), in eV
#@njit
def theta(hnu_tab_aux, Z, a):
 
    hnu_pet_result = hnu_pet(Z, a)

    if Z >= 0:
        return hnu_tab_aux - hnu_pet_result + (Z + 1.0) * q**2 / a / eV
    else:
        return hnu_tab_aux - hnu_pet_result
    
# Equation (11) and preceding paragraph
#@njit
def y2(hnu_tab_aux, Z, a):
    Nnu = len(hnu_tab_aux)
    y2_tab = np.zeros(Nnu)

    if Z >= 0:
        E_low = -(Z + 1.0) * q**2 / a / eV
        E_high = hnu_tab_aux - hnu_pet(Z, a)
        ind = np.where(E_high > max([0.0, E_low]))
        count = len(ind[0])
        if count != 0:
            E_high = E_high[ind]
            y2_tab[ind] = E_high**2 * (E_high - 3.0 * E_low) / (E_high - E_low)**3

    else:
        ind = np.where(hnu_tab_aux > hnu_pet(Z, a))
        count = len(ind[0])
        if count != 0:
            y2_tab[ind] = 1.0

    return y2_tab

def l_a():
    """
    Stores la(hnu) needed for equation (15) of WD01b.    
    Uses tabulated values for Im(m_para) and
    Im(m_perp), taken from Draine website and given for a = 100A, T = 20K.
    We assume the changes due to grain size and temperature are negligible.
    """

    # Read data from files
    fileperp = SpDust_data_dir + 'perpendicular.out'
    filepara = SpDust_data_dir + 'parallel.out'

    lamb_perp, Imm_perp = np.loadtxt(fileperp, usecols=(0, 4), unpack=True, comments=';')
    lamb_para, Imm_para = np.loadtxt(filepara, usecols=(0, 4), unpack=True, comments=';')
   
    # There are more wavelength values given for m_para to account for line effects,
    # so we interpolate Im(m_perp) at those values.

    # Interpolate Im(m_perp) at additional wavelength points
    #Imm_perp_interp = np.exp(np.interp(np.log(lamb_para), np.log(lamb_perp), np.log(Imm_perp)))
    
    #interp_function = interp1d(np.log(lamb_perp), np.log(Imm_perp), kind='cubic', fill_value="extrapolate")
    interp_function = interp1d(np.log(lamb_perp), np.log(Imm_perp), kind='linear', fill_value="zero")
    Imm_perp_interp = np.exp(interp_function(np.log(lamb_para)))

    # Convert wavelength to cm and calculate energy
    lamb_tab = lamb_para * 1e-4  # in cm
    hnu_tab_aux = h * c / (lamb_tab * eV)

    # Calculate la_tab using WD01 (15)
    la_tab_aux = 3 * lamb_tab / (4 * pi) / (2 * Imm_perp_interp + Imm_para)

    print('l_a computed')

    return hnu_tab_aux, la_tab_aux

hnu_tab, la_tab = l_a()
refr_indices.hnu_tab = hnu_tab
refr_indices.la_tab = la_tab 
 
# Takes care of small beta case in eq (13)
#@njit
def fy1(x):
    """
    Takes care of small beta case as described in equation (13).
    
    Parameters:
    - x: NumPy array of values for which the function is computed.
    
    Returns:
    - NumPy array with the computed result for each value of `x`.
    """

    xmin = 1e-5  # Threshold for small x values
    result = np.zeros_like(x)  # Initialize result array with the same shape as x

    # Boolean indexing for small and rest values
    ind_small = x < xmin
    rest = x >= xmin

    # Apply the formula for small x
    if np.any(ind_small):
        result[ind_small] = x[ind_small] / 3.0

    # Apply the formula for the rest of the values
    if np.any(rest):
        xrest = x[rest]
        result[rest] = (xrest**2 - 2.0 * xrest + 2.0 - 2.0 * np.exp(-xrest)) / xrest**2

    return result

# Equation (13) in WD01b
#@njit
def y1(a):
    l_e = 1e-7  # electron escape length in cm

    beta = a / la_tab
    alpha = a / la_tab + a / l_e

    return fy1(alpha) / fy1(beta)

# Equation (16) in WD01b
#@njit
def y0(theta_val):
    return 9e-3 * (theta_val / W)**5 / (1 + 3.7e-2 * (theta_val / W)**5)

# Equation (12) for hnu values in hnu_tab
#@njit
def Y(Z, a):
    N_hnu = len(hnu_tab)
    Y_result = np.zeros(N_hnu)

    theta_values = theta(hnu_tab, Z, a)
    y0_values = y0(theta_values)
    y1_values = y1(a)

    y0y1_values = y0_values * y1_values
    #ind = np.where(y0y1_values > 1.0)
    #if len(ind[0]) > 0:
    #    y0y1_values[ind] = 1.0
    ind = y0y1_values > 1.0
    if np.any(ind):
        y0y1_values[ind] = 1.0


    Y_result = y2(hnu_tab, Z, a) * y0y1_values

    return Y_result

# Planck function for nu in Hz 
#@njit
def Planck_B(nu, T):
    return 2.0 * h * nu**3 / c**2 / (np.exp(h * nu / (k * T)) - 1.0)

#@njit
def nu_uisrf(hnu_tab_aux):
    """
    ; ---> Average interstellar radiation field spectrum nu*u_nu
    ; (Mathis, Mezger & Panagia 1983) see eq (31) of WD01b
    ; IMPORTANT note: the ambiant radiation field is always supposed to
    ; be a multiple of u_ISRF. Of course you may change that but you will
    ; need to re-run some calculations for the Infrared damping and
    ; excitation coefficients (see readme file). 
    """

    # Initialize array
    uisrf_tab = np.zeros(hnu_tab_aux.shape, dtype=np.float64)

    # Conditions for different ranges
    #ind1 = np.where((hnu_tab_aux < 13.6) & (hnu_tab_aux > 11.2))
    ind1 = (hnu_tab_aux < 13.6) & (hnu_tab_aux > 11.2)
    if np.any(ind1):
        uisrf_tab[ind1] = 3.328e-9 * hnu_tab_aux[ind1]**(-4.4172)

    #ind2 = np.where((hnu_tab_aux < 11.2) & (hnu_tab_aux > 9.26))
    ind2 = (hnu_tab_aux < 11.2) & (hnu_tab_aux > 9.26)
    if np.any(ind2):
        uisrf_tab[ind2] = 8.463e-13 / hnu_tab_aux[ind2]

    #ind3 = np.where((hnu_tab_aux < 9.26) & (hnu_tab_aux > 5.04))
    ind3 = (hnu_tab_aux < 9.26) & (hnu_tab_aux > 5.04)
    if np.any(ind3):
        uisrf_tab[ind3] = 2.055e-14 * hnu_tab_aux[ind3]**0.6678

    #ind4 = np.where(hnu_tab_aux < 5.04)
    ind4 = (hnu_tab_aux < 5.04)
    if np.any(ind4):
        nu = hnu_tab_aux[ind4] * eV / h
        uisrf_tab[ind4] = 4 * pi * nu / c * (
            1e-14 * Planck_B(nu, 7500) + 1.65e-13 * Planck_B(nu, 4000) + 4e-13 * Planck_B(nu, 3000)
        )

    return uisrf_tab

def readPAH():

    # File paths
    PAHion = SpDust_data_dir + 'PAHion_read.out'
    PAHneu = SpDust_data_dir + 'PAHneu_read.out'

    # Common variables
    Nrad = 30
    Nwav = 1201

    # Read PAH ion and neutral data
    C1aux, C2_ion = np.loadtxt(PAHion, usecols=(0, 2), unpack=True, comments=';')
    C2_neu = np.loadtxt(PAHneu, usecols=(2,), unpack=True, comments=';')

    # Initialize arrays
    a_tab = np.zeros(Nrad)
    lambda_tab = np.zeros(Nwav)

    Q_abs_neu = np.zeros((Nrad, Nwav))
    Q_abs_ion = np.zeros((Nrad, Nwav))

    # Populate absorption efficiency arrays
    for i in range(Nrad):
        a_tab[i] = C1aux[(Nwav+1)*i] * 1e-4
        for j in range(Nwav):
            lambda_tab[j] = C1aux[(Nwav + 1) * i + j + 1]
            Q_abs_neu[i, j] = C2_neu[(Nwav + 1) * i + j + 1]
            Q_abs_ion[i, j] = C2_ion[(Nwav + 1) * i + j + 1]

    # Calculate wavelength array
    Qabs_hnu_tab = h * c / (lambda_tab * 1e-4) / eV

    print('readPAH computed')

    Qabstabs.Nrad = Nrad
    Qabstabs.Nwav = Nwav
    Qabstabs.a_tab = a_tab
    Qabstabs.Qabs_hnu_tab = Qabs_hnu_tab
    Qabstabs.Q_abs_ion = Q_abs_ion
    Qabstabs.Q_abs_neu = Q_abs_neu
    return 

readPAH()

#Qabs_interp_func_neu = biinterp_func(Qabstabs.Q_abs_neu, np.log(Qabstabs.a_tab), np.log(Qabstabs.Qabs_hnu_tab))
#Qabs_interp_func_ion = biinterp_func(Qabstabs.Q_abs_ion, np.log(Qabstabs.a_tab), np.log(Qabstabs.Qabs_hnu_tab)) 

def Qabs(a, Z, hnu_tab_aux):
    Qabs_hnu_tab = Qabstabs.Qabs_hnu_tab
    a_tab = Qabstabs.a_tab
    Q_abs_ion = Qabstabs.Q_abs_ion
    Q_abs_neu = Qabstabs.Q_abs_neu

    # Minimum frequency for which Q_abs is tabulated
    hnu_min = np.min(Qabs_hnu_tab)

    N_nu = len(hnu_tab_aux)
    Qtab = np.zeros(N_nu)

    # Find indices where hnu_tab > hnu_min
    ind = hnu_tab_aux > hnu_min
    pts = coord_grid(np.log(a), np.log(hnu_tab_aux[ind]))
    if np.any(ind):
        if Z == 0:
            #Qtab[ind] = Qabs_interp_func_neu(pts)[0]
            Qtab[ind] = log_biinterplog(Q_abs_neu, a_tab, Qabs_hnu_tab, a, hnu_tab_aux[ind])[0,:]
        else:
            #Qtab[ind] = Qabs_interp_func_ion(pts)[0]
            Qtab[ind] = log_biinterplog(Q_abs_ion, a_tab, Qabs_hnu_tab, a, hnu_tab_aux[ind])[0,:]

    # For long wavelengths where hnu_tab <= hnu_min, assume Qabs ∝ ν^2
    ind = hnu_tab_aux <= hnu_min
    #pts = coord_grid(np.log(a), np.log(hnu_min))
    if np.any(ind):
        if Z == 0:
            #Qtab[ind] = (hnu_tab_aux[ind] / hnu_min)**2 * Qabs_interp_func_neu(pts)[0]
            Qtab[ind] = (hnu_tab_aux[ind] / hnu_min)**2 * log_biinterplog(Q_abs_neu, a_tab, Qabs_hnu_tab, a, hnu_min)[0,0]
        else:
            #Qtab[ind] = (hnu_tab_aux[ind] / hnu_min)**2 * Qabs_interp_func_ion(pts)[0]
            Qtab[ind] = (hnu_tab_aux[ind] / hnu_min)**2 * log_biinterplog(Q_abs_ion, a_tab, Qabs_hnu_tab, a, hnu_min)[0,0]

    return Qtab

# Equation (4) in eV for Z < 0
#@njit
def EA(Z, a):
    return W + ((Z - 0.5) * q**2 / a - q**2 / a * 4e-8 / (a + 7e-8)) / eV

# Equation (18) in eV for Z < 0
#@njit
def hnu_pdt(Z, a):
    return max(0, EA(Z + 1, a) + E_min(Z, a))  

# Equation (20) in cm^2, for Z < 0
#@njit
def sigma_pdt(hnu_tab_aux, Z, a):
    DeltaE = 3.0  # in eV
    x = (hnu_tab_aux - hnu_pdt(Z, a)) / DeltaE
    sigma = np.zeros_like(hnu_tab_aux)

    if Z >= 0:
        return sigma

    ind = x > 0
    if np.any(ind):
        x = x[ind]
        sigma[ind] = 1.2e-17 * np.abs(Z) * x / (1 + x**2 / 3)**2

    return sigma

# First term in eq (25) for the standard interstellar radiation field (31)


def first_term(Z, a):
    hnu_min = max([1e-5, hnu_pet(Z, a)])
    hnu_max = 13.6

    # Parameter
    Nnu = 500

    if hnu_min > hnu_max:
        return 0.0

    hnu = makelogtab(hnu_min, hnu_max, Nnu)
    Dnu_over_nu = DX_over_X(hnu_min, hnu_max, Nnu)

    Ytab = np.interp(np.log(hnu), np.log(hnu_tab), Y(Z, a))

    Qtab = Qabs(a, Z, hnu)
    utab = nu_uisrf(hnu)

    return c * pi * a**2 * Dnu_over_nu * np.sum(Ytab * Qtab * utab / (hnu * eV))

# Second term in eq (25) for the standard interstellar radiation field (31), for Z < 0
#@njit
def second_term(Z, a):

    hnu_min = max([1e-5, hnu_pdt(Z, a)])
    hnu_max = 13.6

    # Parameter
    Nnu = 500

    if Z >= 0 or hnu_pdt(Z, a) > hnu_max:
        return 0.0

    hnu = makelogtab(hnu_min, hnu_max, Nnu)
    Dnu_over_nu = DX_over_X(hnu_min, hnu_max, Nnu)

    utab = nu_uisrf(hnu)
    sigmatab = sigma_pdt(hnu, Z, a)

    return c * np.sum(sigmatab * utab / (hnu * eV)) * Dnu_over_nu

# Equation (21)
def Jpe(Z, a):
    return first_term(Z, a) + second_term(Z, a)

# Equation (22)
#@njit
def Zmax(a):
    aA = a / 1e-8  # a in A

    hnu_max = 13.6  # You may need to change this value
    result= np.floor(((hnu_max - W)/ 14.4 * aA + 0.5 - 0.3 / aA) / (1 + 0.3 / aA))
    return int(result)

# Equations (23), (24)
#@njit
def Zmin(a):
    aA = a / 1e-8  # a in A
    U_ait = -(3.9 + 0.12 * aA + 2 / aA)
    result = np.floor(U_ait / 14.4 * aA) + 1
    return int(result)

# Function to compute and store Jpeisrf
def JPEisrf_calc(filename='jpeisrf_data.npz'):
    """
    ; ---> Computes and stores the photoemission rate Jpe(a, Z), for an
    ; array of grain radii and charges, for tha averge interstellar
    ; radiation field.

    If previously computed results exist, they will be loaded.
    """
    # Check if the saved file exists
    filename = os.path.join(SpDust_data_dir, filename)
    if os.path.exists(filename):
        # Load saved data
        data = np.load(filename)
        a_values = data['a_values']
        Jpe_pos_isrf = data['Jpe_pos_isrf']
        Jpe_neg_isrf = data['Jpe_neg_isrf']
        print('Loaded previously computed Jpeisrf arrays from file.')
    else:
        a_min = 3.5e-8
        a_max = 1e-6
        Na = 30
        # Initialize arrays
        a_values = makelogtab(a_min, a_max, Na)
        Z_min = Zmin(a_max)
        Z_max = Zmax(a_max)
        Jpe_neg_isrf = np.zeros((Na, abs(Z_min) + 1))
        Jpe_pos_isrf = np.zeros((Na, Z_max + 1))

        # Loop to compute Jpeisrf
        for i in range(Na):
            a = a_values[i]
            for j in range(Zmax(a) + 1):
                Z = j
                Jpe_pos_isrf[i, j] = Jpe(Z, a)
                if j <= abs(Zmin(a)):
                    Z = -j
                    Jpe_neg_isrf[i, j] = Jpe(Z, a)

        print('Jpeisrf computed')

        # Save computed arrays to file for future use
        np.savez(filename, a_values=a_values, Jpe_pos_isrf=Jpe_pos_isrf, Jpe_neg_isrf=Jpe_neg_isrf)
        print(f'Jpeisrf arrays saved to {filename}.')

    jpe_arrays.a_values = a_values
    jpe_arrays.Jpe_pos_isrf = Jpe_pos_isrf
    jpe_arrays.Jpe_neg_isrf = Jpe_neg_isrf

    return a_values, Jpe_pos_isrf, Jpe_neg_isrf

# Call the function
a_values, Jpe_pos_isrf, Jpe_neg_isrf = JPEisrf_calc()


# Function to interpolate Jpe arrays
#@jit
def JPEisrf(a):
    #a_values, Jpe_pos_isrf, Jpe_neg_isrf = jpe_arrays.a_values, jpe_arrays.Jpe_pos_isrf, jpe_arrays.Jpe_neg_isrf

    Jpepos = np.zeros(len(Jpe_pos_isrf[0, :]))
    Jpeneg = np.zeros(len(Jpe_neg_isrf[0, :]))

    if a <= np.min(a_values):
        Jpepos = Jpe_pos_isrf[0, :]
        Jpeneg = Jpe_neg_isrf[0, :]
    else:
        if a >= np.max(a_values):
            N_a = len(a_values)
            Jpepos = Jpe_pos_isrf[N_a - 1, :]
            Jpeneg = Jpe_neg_isrf[N_a - 1, :]
        else:
            #ia = np.max(np.where(a_values <= a))
                # Use np.where with index extraction
            indices = np.where(a_values <= a)[0]
            if indices.size > 0:
                ia = indices[-1]  # Take the last index where the condition holds
            else:
                ia = 0  # Handle case where no values match the condition
            alpha = np.log(a / a_values[ia]) / np.log(a_values[ia + 1] / a_values[ia])
            Jpepos = (1.0 - alpha) * Jpe_pos_isrf[ia, :] + alpha * Jpe_pos_isrf[ia + 1, :]
            Jpeneg = (1.0 - alpha) * Jpe_neg_isrf[ia, :] + alpha * Jpe_neg_isrf[ia + 1, :]

    Zmin_val = Zmin(a)
    Zmax_val = Zmax(a)

    Jpepos = Jpepos[:Zmax_val + 1]
    Jpeneg = Jpeneg[:-Zmin_val + 1]

    return Jpepos, Jpeneg

# Function to calculate Jtilde
#@njit
def Jtilde(tau, nu):

    if nu == 0:
        return 1.0 + np.sqrt(pi / (2.0 * tau))  # Equation (3.3)

    if nu < 0:
        return (1.0 - nu / tau) * (1.0 + np.sqrt(2.0 / (tau - 2.0 * nu)))  # Equation (3.4)

    if nu > 0:
        ksi = 1.0 + 1.0 / np.sqrt(3.0 * nu)
        theta = nu / ksi - 0.5 / (ksi**2 * (ksi**2 - 1.0))  # Equation (2.4a)

        if theta / tau < 700.0:  # Avoid floating-point underflow
            return (1.0 + (4.0 * tau + 3.0 * nu)**(-0.5))**2 * np.exp(-theta / tau)  # Equation (3.5)
        else:
            return 0.0
        
# Function to calculate the sticking coefficient for electrons
# Equations refer to WD01b
# Note that s_i = 1 for ions

#@njit
def se(Z, a):
    l_e = 1e-7  # "electron escape length" (10 A) see below (28)
    Nc = N_C(a)
    Z_min = Zmin(a)  

    if Z == 0 or (Z < 0 and Z > Z_min):
        return 0.5 * (1.0 - np.exp(-a / l_e)) / (1.0 + np.exp(20.0 - Nc))  # (28),(29)

    if Z < 0 and Z <= Z_min:
        return 0.0  # (29)

    if Z > 0:
        return 0.5 * (1.0 - np.exp(-a / l_e))  # (30)

# Function to calculate J_ion
#@njit
def J_ion_aux(nh, T, xh, xc, Z, a):
    asurf_val = asurf(a)  
    tau = asurf_val * k * T / q**2
    nu = Z 
    return nh * np.sqrt(8.0 * k * T / (pi * mp)) * pi * asurf_val**2 * (xh + xc / np.sqrt(12.0)) * Jtilde(tau, nu)

def J_ion(env, Z, a):
    nh = env['nh']
    T = env['T']
    xh = env['xh']
    xc = env['xC']
    return J_ion_aux(nh, T, xh, xc, Z, a)

# Function to calculate J_electron
#@njit
def J_electron_aux(nh, T, xh, xc, Z, a):
    asurf_val = asurf(a)  
    tau = asurf_val * k * T / q**2
    nu = -Z 
    return nh * (xh + xc) * se(Z, a) * np.sqrt(8.0 * k * T / (np.pi * me)) * pi * asurf_val**2 * Jtilde(tau, nu)

def J_electron(env, Z, a):
    nh = env['nh']
    T = env['T']
    xh = env['xh']
    xc = env['xC']
    return J_electron_aux(nh, T, xh, xc, Z, a) 

# Function to compute charge distribution
#@jit
def charge_dist_aux(Chi, nh, T, xh, xc, a):
    Z_min = Zmin(a) 
    Z_max = Zmax(a)  

    fneg = np.zeros(abs(Z_min) + 1)
    fpos = np.zeros(Z_max + 1)

    Ji_neg = np.zeros(abs(Z_min) + 1)
    Je_neg = np.zeros(abs(Z_min) + 1)
    Ji_pos = np.zeros(Z_max + 1)
    Je_pos = np.zeros(Z_max + 1)
    Jpepos_val, Jpeneg_val = JPEisrf(a)

    for j in range(abs(Z_min) + 1):
        Z = -j
        Ji_neg[j] = J_ion_aux(nh, T, xh, xc, Z, a)
        Je_neg[j] = J_electron_aux(nh, T, xh, xc, Z, a)
    Jpe_neg = Chi * Jpeneg_val

    for Z in range(Z_max + 1):
        Ji_pos[Z] = J_ion_aux(nh, T, xh, xc, Z, a)
        Je_pos[Z] = J_electron_aux(nh, T, xh, xc, Z, a)
    Jpe_pos = Chi * Jpepos_val

    # Compute charge distribution function using DL98b (4)
    fpos[0] = 1.0
    fneg[0] = 1.0

    Z = -1
    while Z >= Z_min:
        fneg[-Z] = fneg[-(Z + 1)] * Je_neg[-(Z + 1)] / (Ji_neg[-Z] + Jpe_neg[-Z])
        Z -= 1
        fneg /= np.sum(fneg)

    Z = 1
    while Z <= Z_max:
        fpos[Z] = fpos[Z - 1] * (Ji_pos[Z - 1] + Jpe_pos[Z - 1]) / Je_pos[Z]
        Z += 1
        fpos /= np.sum(fpos)

    if fneg[0] > 0:
        fneg = fpos[0] / fneg[0] * fneg
    else:
        fpos = np.zeros(1)

    norm = np.sum(fneg) + np.sum(fpos) - fpos[0]

    fneg /= norm
    fpos /= norm

    Zneg = -np.arange(abs(Z_min) + 1)
    Zpos = np.arange(Z_max + 1)

    # Keep only the values for which f_Z > fmin
    ind_neg = np.where((fneg > 1e-10) & (Zneg != 0))
    if len(ind_neg[0]) > 0:
        fneg = fneg[ind_neg]
        Zneg = Zneg[ind_neg]

    ind_pos = np.where((fpos > 1e-10) | (Zpos == 0))
    fpos = fpos[ind_pos]
    Zpos = Zpos[ind_pos]

    # Combine arrays
    fZ = np.zeros((2, len(Zpos) + len(Zneg)))

    fZ_array = fpos
    Z_array = Zpos

    if len(Zneg) > 0:
        Z_array = np.concatenate((Z_array, Zneg))
        fZ_array = np.concatenate((fZ_array, fneg))

    # Renormalize
    fZ_array /= np.sum(fZ_array)

    fZ[0, :] = Z_array
    fZ[1, :] = fZ_array

    return fZ

def charge_dist(env, a):
    Chi = env['Chi']
    nh = env['nh']
    T = env['T']
    xh = env['xh']
    xc = env['xC']

    return charge_dist_aux(Chi, nh, T, xh, xc, a)



from .util import cgsconst, makelogtab, DX_over_X
from scipy.special import erf
from . import SpDust_data_dir
import numpy as np
import os

class grainparams:
    # Stores some grain geometrical properties
    a2 = 6e-8          # radius under which grains are clyindrical, above which grains are ellipsoidal
    d = 3.35e-8        # assumed minimal disk thickness (graphite interlayer separation)
    rho = 2.24         # carbon density in g/cm^3
    epsilon = 0.01     # charge centroid displacement

rho = grainparams.rho # carbon density in g/cm^3
a2 = grainparams.a2
d = grainparams.d
epsilon = grainparams.epsilon

pi = np.pi
mp = cgsconst.mp
q = cgsconst.q

# Number of carbon atoms in a grain
#@njit
def N_C(a):
    return int(np.floor(4 * pi / 3 * a**3 * rho / (12 * mp)) + 1)

# Number of hydrogen atoms in a grain
#@njit
def N_H(a):
    Nc = N_C(a)
    if Nc < 25:
        result = np.floor(0.5 * Nc + 0.5)
    elif Nc < 100:
        result = np.floor(2.5 * np.sqrt(Nc) + 0.5)
    else:
        result = np.floor(0.25 * Nc + 0.5)
    return int(result)
    
#@njit
def Inertia_z(I_ref, beta):
    """
    Calculate the moment of inertia around the z-axis. Also notated as I_3.
    """
    return I_ref / (1 + beta)
    
#@njit    
def grain_mass(a):
    """"
    Calculate the mass of the grain given the effective radius. (Functions below will define the effective radius)

    Parameters:
    -----------
    a: float
        The effective radius of the grain.
    """
    return (12 * N_C(a) + N_H(a)) * mp

########################################### Functions for cylindrical type grains ###########################################
#@njit
def effective_radius_cylindrical(I_ref, beta):
    '''
    Calculate the effective radius of an elliptical cylinder grain.
    '''
    result = (9 / 4*pi*rho) * I_ref / (1 + beta) * (beta + 0.5)**(1/3)
    return result**(1/5)

#@njit
def Inertia_ref_cylindrical(a, beta):
    '''
    Calculate the reference moment of inertia.
    '''
    aux = 9 * (beta + 0.5)**(1/3) / (1 + beta)
    mass = grain_mass(a)
    real_rho = mass / (4/3 * pi * a**3)
    return a**5 * 4 * pi * real_rho / aux 

#@njit
def cylindrical_radius(a, beta):
    '''
    Calculate the radius of the cylindrical grain in the plane spaned by grain body axes 1 and 2.
    '''
    # a = effective_radius_cylindrical(I_ref, beta)
    mass = grain_mass(a)
    I_ref = Inertia_ref_cylindrical(a, beta)
    return np.sqrt( 2 * Inertia_z(I_ref, beta) / mass )

#@njit
def cylindrical_thickness(a, beta):
    """
    Calculate the thickness of the cylindrical grain.
    """

    mass = grain_mass(a)
    I_ref = Inertia_ref_cylindrical(a, beta)
    result = (2 * I_ref - Inertia_z(I_ref, beta)) * 6 / mass
    return np.sqrt(result)

#@njit
def cylindrical_params(a, d_val):
    '''
    Given the effective radius (a) and thickness (d), calculate the I_ref and beta
    '''
    mass = grain_mass(a)
    b_squared = 4/3 * a**3 / d_val
    I_ref = (b_squared / 4 + d_val**2 / 12) * mass
    I_z = 0.5 * b_squared  * mass
    beta = I_ref / I_z - 1
    return I_ref, beta

#@njit
def beta_min(a):
    '''
    Calculate the minimum value of beta for a cylindrical grain, assuming the thinest disk-like grain - single .
    '''
    return cylindrical_params(a, d)[1]

#@njit
def asurf_cylindrical(a, beta):
    ''''
    Calculate the surface-equivalent radius for cylindrical grains.
    '''
    b_val = cylindrical_radius(a, beta)
    d_val = cylindrical_thickness(a, beta)
    return np.sqrt(b_val**2 / 2 + b_val * d_val / 2)

#@njit
def acx_cylindrical(a, beta):
    '''
    Calculate the cylindrical excitation equivalent radius for cylindrical grains. Assume thin disk-like grains.
    '''
    return (3 / 8)**0.25 * cylindrical_radius(a, beta)


########################################### Functions for ellipsoid grains ###########################################
#@njit
def effective_radius_ellipsoidal(I_ref, beta):
    '''
    Calculate the effective radius of an ellipsoidal grain.
    '''
    result = (15 * I_ref / 8*pi*rho) / (1 + beta) * (2 * beta + 1)**(1/3)
    return result**(1/5)

#@njit
def Inertia_ref_ellipsoidal(a, beta):
    '''
    Calculate the reference moment of inertia.
    '''
    aux = 15 * (2 * beta + 1)**(1/3) / (1 + beta)
    mass = grain_mass(a)
    real_rho = mass / (4/3 * pi * a**3)
    return a**5 * 8 * pi * real_rho / aux

#@njit
def radii_ellipsoidal(a, beta):
    '''
    Calculate the radii of the ellipsoidal grain.
    '''
    I_ref = Inertia_ref_ellipsoidal(a, beta)
    I1 = I2 = I_ref
    I3 = Inertia_z(I_ref, beta)

    mass = grain_mass(a)

    r_1 = r_2 = np.sqrt(5 * I3 / (2 * mass))
    r_3 = np.sqrt(5 * (I1 + I2 - I3) / (2 * mass))

    return r_1, r_2, r_3

#@njit
def asurf_ellipsoidal(a, beta):
    ''''
    Calculate the surface-equivalent radius for ellipsoidal grains.
    '''
    if beta == 0: # Spherical grain
        return a
    elif beta < 0: # Oblate grain
        r_1, r_2, r_3 = radii_ellipsoidal(a, beta)
        e = np.sqrt(1 - r_3**2 / r_1**2)
        surface = 2 * pi * r_1**2 * (1 + (1 - e**2) / e * np.arctanh(e))
        asurf = np.sqrt(surface / (4 * pi))
        return asurf
    else: # Prolate grain
        r_1, r_2, r_3 = radii_ellipsoidal(a, beta)
        e = np.sqrt(1 - r_1**2 / r_3**2)
        surface = 2 * pi * r_1**2 * (1 + r_3 / r_1 / e * np.arcsin(e))
        asurf = np.sqrt(surface / (4 * pi))
        return asurf

#@njit
def acx_ellipsoidal(a, beta):
    '''
    Calculate the excitation equivalent radius for ellipsoidal grains.

    Assume nearly spherical grains; use asurf as the excitation equivalent radius.
    '''
    return asurf_ellipsoidal(a, beta) 

########################################### Functions for general grains ###########################################
# Assume cylindrical grains for a < a2 and ellipsoidal grains for a > a2

#@njit
def Inertia_ref(a, beta):
    '''
    Calculate the reference moment of inertia.
    '''
    if a <= a2:
        return Inertia_ref_cylindrical(a, beta)
    else:
        return Inertia_ref_ellipsoidal(a, beta)
    
#@njit
def asurf(a, beta):
    '''
    Calculate the surface-equivalent radius for general grains.
    '''
    if a <= a2:
        return asurf_cylindrical(a, beta)
    else:
        return asurf_ellipsoidal(a, beta)
    
#@njit
def acx(a, beta):
    '''
    Calculate the excitation equivalent radius for general grains.
    '''
    if a <= a2:
        return acx_cylindrical(a, beta)
    else:
        return acx_ellipsoidal(a, beta)
    
#@njit
def Inertia_largest(a, beta):
    '''
    Calculate the largest moment of inertia.
    '''

    I_ref = Inertia_ref(a, beta)

    if beta < 0:
        return Inertia_z(I_ref, beta)
    else:
        return I_ref
    
#@njit
def rms_dipole(a, beta, Z2, mole_dipole):
    '''
    Calculate the root mean square dipole moment.

    Parameters:
    -----------
    a: float
        The effective radius of the grain.
    Z2: float
        The square of the charge.
    mole_dipole: float
        The typical molecular dipole moment. (This is equivalent to the "beta" in spdust (AHD09;SAH11). Don't confuse with the grain parameter beta in SpyDust.)
    '''
    muZ = epsilon * np.sqrt(Z2) * q * acx(a, beta)
    N_at = N_C(a) + N_H(a)
    return np.sqrt(N_at * mole_dipole**2 + muZ**2)

########################################### Functions for grain size and shape distributions ###########################################

# Define the path to the size distribution file
size_dist_file = os.path.join(SpDust_data_dir, 'sizedists_table1.out')

class size_dist_arrays:
    bc1e5_tab, alpha_tab, beta_tab, at_tab, ac_tab, C_tab = \
    np.loadtxt(size_dist_file, usecols=(1, 3, 4, 5, 6, 7), unpack=True, comments=';')

#@njit
def apodized_Gaussian_dist(x, centre, sigma, right_half = False):
    '''
    An auxiliary Gaussian distribution apodized by a cosine function.
    This is a finite distribution with range [centre - 2 * sigma, centre + 2 * sigma].
    '''
    if x >= centre + 2 * sigma:
        return 0
    elif x>= centre:
        result = np.exp(-0.5 * (x - centre)**2 / sigma**2) / (sigma * np.sqrt(2 * np.pi)) # Gaussian part
        result *= np.cos(np.pi * (x - centre) / 4 / sigma) # Apodization at 2 sigma
        return result   
    elif right_half or x <= centre - 2 * sigma:
        return 0
    else:
        result = np.exp(-0.5 * (x - centre)**2 / sigma**2) / (sigma * np.sqrt(2 * np.pi))
        result *= np.cos(np.pi * (x - centre) / 4 / sigma)
        return result

class grain_distribution():
    bc, alpha_g, beta_g, at_g, ac_g, C_g = None, None, None, None, None, None

    def __init__(self, a_min=3.5e-8, a_max=3.5e-7, Na=30): # Default settings for grain size distribution
        self.a_min = a_min
        self.a_max = a_max
        self.Na = Na
        self.a_tab = makelogtab(a_min, a_max, Na)
        self.Da_over_a = DX_over_X(a_min, a_max, Na)

    def size_dist_func(self, line): 
        # Unpack size distribution parameters
        bc = size_dist_arrays.bc1e5_tab[line] * 1e-5
        alpha_g = size_dist_arrays.alpha_tab[line]
        beta_g = size_dist_arrays.beta_tab[line]
        at_g = size_dist_arrays.at_tab[line] * 1e-4
        ac_g = size_dist_arrays.ac_tab[line] * 1e-4
        C_g = size_dist_arrays.C_tab[line]

        amin = self.a_min
        # Lognormal populations parameters
        mc = 12 * mp
        bci = np.array([0.75, 0.25]) * bc
        a0i = np.array([3.5, 30.]) * 1e-8
        sigma = 0.4

        Bi = 3 / (2 * np.pi)**1.5 * np.exp(-4.5 * sigma**2) / (
                rho * a0i**3 * sigma) * bci * mc / (1 + erf(3 * sigma / np.sqrt(2) + np.log(a0i / amin) / (sigma * np.sqrt(2))))
        
        #@njit
        def size_dist(a):
            '''
            Grain size distribution, using Weingartner & Draine, 2001a prescription. 
            The line of their table 1 is given by the user in param_file.
            '''
            
            D_a = np.sum(Bi / a * np.exp(-0.5 * (np.log(a / a0i) / sigma)**2))

            if beta_g >= 0:
                F_a = 1 + beta_g * a / at_g
            else:
                F_a = 1 / (1 - beta_g * a / at_g)

            cutoff = 1
            if a > at_g:
                cutoff = np.exp(-((a - at_g) / ac_g)**3)

            return D_a + C_g / a * (a / at_g)**alpha_g * F_a * cutoff
        
        return size_dist

    @staticmethod
    #@njit
    def shape_dist(a, beta_tab):
        '''
        A toy model for the distribution of the grain shape parameter, beta. 
        Note that this is a conditional distribution, given the grain size.
        We assume a Gaussian distribution apodized by a cosine function.
        '''
        if a <= a2:
            beta_min = cylindrical_params(a, d)[1]
            sigma = 0.1
            # Calculate the distribution
            dist = np.zeros_like(beta_tab)
            for i, beta in enumerate(beta_tab):
                dist[i] = apodized_Gaussian_dist(beta, beta_min, sigma, right_half=True)
            # Normalize the distribution
            dist /= np.sum(dist)
            return dist
        else:
            sigma = 0.025
            dist = np.zeros_like(beta_tab)
            for i, beta in enumerate(beta_tab):
                dist[i] = apodized_Gaussian_dist(beta, 0, sigma)
            dist /= np.sum(dist)
            return dist

    
    @staticmethod
    #@njit 
    def shape_dist_fixed_thickness(a, beta_tab):
        '''
        The distribution of the grain shape parameter, beta, if we fix the thickness of the disk as d. 
        Note that this is a conditional distribution, given the grain size.
        We assume a Gaussian distribution apodized by a cosine function.
        '''
        if a <= a2:
            beta = cylindrical_params(a, d)[1]
            dist = np.zeros_like(beta_tab)
            # find the index of the element closest to beta in beta_tab
            idx = np.abs(beta_tab - beta).argmin()
            dist[idx] = 1
            return dist
        else:
            dist = np.zeros_like(beta_tab)
            # find the index of the element closest to 0 in beta_tab
            idx = np.abs(beta_tab).argmin()
            dist[idx] = 1
            return dist

    def shape_and_size_dist(self, line, a_weighted=True, normalize=False, Nbeta=50, fixed_thickness=False):
        '''
        Calculate the joint distribution of the grain shape parameter, beta, and the grain size, a.

        Parameters:
        -----------
        line: int
            The line of the size distribution table.
        
        a_weighted: bool
            If True, the grain size distribution is weighted and we get f(a)da;
            otherwise, we get f(a).

        Returns:
        --------
        A 2D array with the joint distribution.
        '''
        size_func = self.size_dist_func(line)
        beta_tab = np.linspace(-0.49, 0.1, Nbeta)
        result = np.zeros((len(self.a_tab), len(beta_tab)))
        a_dist = np.zeros_like(self.a_tab)
        for ai, a in enumerate(self.a_tab):
            a_dist[ai] = size_func(a)
        if a_weighted:
            a_dist = a_dist * self.Da_over_a * self.a_tab
        if normalize:
            a_dist /= np.sum(a_dist)
        self.a_dist = a_dist

        if fixed_thickness:
            for ai, a in enumerate(self.a_tab):
                result[ai,:] = self.shape_dist_fixed_thickness(a, beta_tab) * a_dist[ai]
        else:
            for ai, a in enumerate(self.a_tab):
                result[ai,:] = a_dist[ai] * self.shape_dist(a, beta_tab)

        aux = np.sum(result, axis=0)
        ind = aux > 1e-500 # Avoid beta values with very small probability

        self.beta_tab = beta_tab[ind]
        return result[:, ind] 

        



import numpy as np
from .util import cgsconst, makelogtab, DX_over_X
from .Grain import acx, Inertia_largest, grainparams
from .infrared import FGIR_averaged
from .collisions import Tev_effective, FGn_averaged, FGi_averaged
from .plasmadrag import FGp_averaged
from .H2_photoemission import GH2, FGpe_averaged

# To compare with the old plasma drag calculation in SpDust
from .SPDUST_as_is import plasmadrag as spd_plasmadrag
from scipy.interpolate import interp1d


c = cgsconst.c
mp = cgsconst.mp
k = cgsconst.k
pi = np.pi

a2 = grainparams.a2



# Function to calculate the characteristic damping time through collisions with neutral H atoms
# @njit
def tau_H(temp, nh, a, beta):
    """
    Returns the characteristic damping time through collisions with neutral H atoms.
    This corresponds to Eq. (24) in AHD09.

    Parameters:
    - env: Environment parameters (temperature T, hydrogen density nh).
    - a: Grain size.
    - temp:   temperature in K
    - nh:  hydrogen density in cm^-3

    Returns:
    - Damping time tau_H (in seconds).
    """

    acx_val = acx(a, beta)  # characteristic length scale
    Inertia_val = Inertia_largest(a, beta)  # grain's moment of inertia

    # Calculate tau_H using the given equation
    return 1.0 / (nh * mp * np.sqrt(2 * k * temp / (pi * mp)) * 4 * pi * acx_val**4 / (3 * Inertia_val))

# @njit
def dissipation_vec_ed( L_val, theta_b, beta, mu_abs, ip, I_ref):
    '''
    Calculate the angular momentum dissipation rate for the back-reaction of the electric dipole radiation.

    Parameters:
    -----------
    L_val: float
        Magnitude of the angular momentum.
    theta_b: float
        The internal angle between the grain body z-axis and the angular momentum.
    beta: float
        The grain parameter, describing the shape of the grain.
    mu_abs: float
        The absolute value of the dipole moment.
    ip: float
        The in-plane ratio of the dipole moment. ip = mu_perp/mu_abs.
    I_ref: float
        The reference moment of inertia.

    Returns:
    --------
        The angular momentum dissipation rate.
    '''

    mu_parallel_2 = mu_abs**2 * (1 - ip**2)
    mu_perp_2 = mu_abs**2 * ip**2

    aux_perp = 1 + ( 1 + 3 * beta * (2 + beta) ) * np.cos(theta_b)**2 + beta**2 * ( 3 + 2 * beta ) * np.cos(theta_b)**4
    aux_parallel = 2 * np.sin(theta_b)**2 

    result = - (1/3) * ( mu_perp_2 * aux_perp + mu_parallel_2 * aux_parallel ) * (L_val / (I_ref * c)) ** 3
    return result

# @njit
def dissipation_rate_ed_theta_avrg( L_val, beta, mu_abs, ip, I_ref):
    '''
    Calculate the ed angular momentum dissipation rate, averaging over nutation angle.
    '''

    mu_perp_2 = mu_abs**2 * ip**2
    mu_parallel_2 = mu_abs**2 * (1 - ip**2)

    aux_perp = (2/5) * beta ** 3 + (8/5) * beta ** 2 + 2 * beta
    aux_parallel = 4 / 3 

    result = - (1/3) * ( mu_perp_2 * aux_perp + mu_parallel_2 * aux_parallel ) * (L_val / (I_ref * c)) ** 3
    return result


# Function to calculate the inverse of the characteristic damping time through electric dipole radiation
# @njit
def tau_ed_inv(temp, a, beta, mu_ip, mu_op, tumbling=True):
    """
    Returns the inverse of the characteristic damping time through electric dipole radiation.
    This corresponds to Eq. (29) in AHD09 and Eq. (47) in SAH10 for tumbling disklike grains.

    Parameters:
    - temp: Environment temperature T in K.
    - a: Grain size.
    - mu_ip: Electric dipole moment in-plane.
    - mu_op: Electric dipole moment out-of-plane.
    - tumbling: Boolean, if True, use the tumbling disk model.

    Returns:
    - Inverse of the damping time (in seconds^-1).
    """  
    
    Inertia_val = Inertia_largest(a, beta)  # grain's moment of inertia
    result = k * temp / (Inertia_val**2 * c**3 * (1+beta)**3)

    if beta==0: # spherical grain
        tumbling = False

    if tumbling:
        # Tumbling disklike grains, SAH10 Eq. (47)
        return result * ((2 * beta**3 / 5 + 8 * beta**2 / 5 + 2*beta + 4/3) * mu_ip**2 + 4 / 3 * mu_op**2) 
    elif beta<=0:
            # oblate grain, theta=0
        return result * (2 * beta**3 + 6*beta**2 + 6*beta + 2 ) * mu_ip**2
    else:
        # prolate grain, theta=pi/2
        return result * (mu_ip**2 + 2 * mu_op**2) 



# Manually implemented cumulative sum function
# @njit
def cumsum_axis_0(arr):
    result = np.zeros_like(arr)
    result[0] = arr[0]
    for i in range(1, arr.shape[0]):
        result[i] = result[i-1] + arr[i]
    return result


def aux_int(Tval, F, G, omega, Nomega, Nmu, tau_H_val, tau_ed_inv_val, Inertia_val, Dln_omega):

     # Rotational distribution function, AHD09 Eq.(33)
    f_a = np.zeros((Nomega, Nmu))

    tau_ed_inv_matrix = np.ones((Nomega, 1)) @ tau_ed_inv_val.reshape(1, -1) # shape (Nomega, Nmu)
    omega_tab = omega.reshape(-1, 1) @ np.ones((1, Nmu)) # shape (Nomega, Nmu)

    X = Inertia_val * omega_tab**2 / (k * Tval)     # shape (Nomega, Nmu)
    integrand = F / G * X + tau_H_val / (3 * G) * tau_ed_inv_matrix * X**2 # shape (Nomega, Nmu)

    #exponent = np.cumsum(integrand, axis=0) * Dln_omega 
    
    #exponent = cumsum_axis_0(integrand)* Dln_omega 
    exponent = np.cumsum(integrand, axis=0) * Dln_omega  # shape (Nomega, Nmu)

    #expo_min = np.min(exponent, axis=0)
    #expo_max = np.max(exponent, axis=0)
    #expo_mid = (expo_max + expo_min) / 2    # shape (Nmu,)
    #print("expo_min, expo_max", expo_min, expo_max)
    #exponent = exponent - expo_mid[np.newaxis, :]
    #exponent = np.clip(exponent, -500, 500)
    #exponent = exponent - (expo_max + expo_min) / 2
    #exponent_aux = 3*np.log(omega_tab) - exponent + np.log(4 * pi * Dln_omega) 
    #norm = np.sum(np.exp(exponent_aux), axis=0) 
    
    # norm = 4 * pi * np.sum(omega_tab**3 * np.exp(-exponent), axis=0) * Dln_omega
    # norm = np.ones((Nomega, 1)) @ norm.reshape(1, -1)
    # f_a = 4 * pi * omega_tab**2 / norm * np.exp(-exponent)

    norm = 4 * pi * np.sum( np.exp(3 * np.log(omega_tab) - exponent), axis=0) * Dln_omega
    log_norm = np.log(norm)

    log_norm = np.ones((Nomega, 1)) @ log_norm.reshape(1, -1)
    
    log_f_a = np.log(4 * pi * omega_tab**2) - exponent - log_norm
    return log_f_a.T   # shape (Nmu, Nomega)


# @njit
def rescale_f_rot(omega, f_a, beta, log=True):
    """
    This function rescales omega in spdust convention to our convention.

    Parameters:
    - omega: array of frequencies.
    - f_a: rotational distribution function (2D array with dimensions [Nmu, Nomega]).
    - log: Boolean, whether the input f_a is in log scale.
    """
    
    omega_new = omega / (1 + beta)
    if log:
        result = f_a + np.log(1+beta)
        return np.log(omega_new), result
    f_a_new =  f_a * (1 + beta)
    return omega_new, f_a_new

    
def log_f_rot(env, a, beta, fZ, mu_ip, mu_op, tumbling=True, omega_min=1e8, omega_max=1e16, Nomega=1000, use_spdust_plasma=False):
    """
    Returns the rotational distribution function f_a:
    f(Omega | a, beta, mu)
    The function is calculated at frequencies centered on the approximate peak frequency
    of the emitted power. The output is a dictionary {omega, f_a} containing:
    - omega: array of frequencies.
    - f_a: rotational distribution function (2D array with dimensions [Nmu, Nomega]).

    Parameters:
    - env: Environmental parameters (T, nh).
    - a: Grain size.
    - ia: Index or identifier for intermediate value saving (optional).
    - fZ: Distribution of grain charges.
    - mu_ip: Electric dipole moment in-plane.
    - mu_op: Electric dipole moment out-of-plane.
    - tumbling: Boolean, whether tumbling disklike grains are considered.

    Returns:
    - Dictionary containing omega and f_a arrays.
    """
    Nmu = np.size(mu_ip)  # Number of dipole moments

    Inertia_val = Inertia_largest(a, beta)  # Grain's moment of inertia

    Tval, nH = env['T'], env['nh']

    if beta==0: # spherical grain
        tumbling = False

    # Characteristic timescales
    tau_H_val = tau_H(Tval, nH, a, beta)
    tau_ed_inv_val = tau_ed_inv(Tval, a, beta, mu_ip, mu_op, tumbling=tumbling)

    # Evaporation temperature
    Tev_val = Tev_effective(env, a)

    # F's and G's (except for plasma drag)
    FGn = FGn_averaged(env, a, beta, Tev_val, fZ, tumbling=tumbling)
    Fn = FGn['Fn']
    Gn = FGn['Gn']

    mu_tot = np.sqrt(mu_ip**2 + mu_op**2)
    FGi = FGi_averaged(env, a, beta, Tev_val, mu_tot, fZ)
    Fi = FGi['Fi']
    Gi = FGi['Gi']

    FGIR = FGIR_averaged(env, a, beta, fZ)
    FIR = FGIR['FIR']
    GIR = FGIR['GIR']

    FGpe = FGpe_averaged(env, a, beta, fZ)
    Fpe = FGpe['Fpe']
    Gpe = FGpe['Gpe']

    GH2_val = GH2(env, a, beta)

    # Array of omegas around the approximate peak
    omega_peak_th = np.array([np.sqrt(6 * k * Tval / Inertia_val)])  # Peak of spectrum if thermal rotation

    # Peak frequency for the lowest and highest values of mu_ip, mu_op
    FGp = FGp_averaged(env, a, beta, fZ, omega_peak_th, [min(mu_ip), max(mu_ip)], [min(mu_op), max(mu_op)], tumbling=tumbling)
    Fp_th = FGp['Fp']
    Gp_th = FGp['Gp']

    F_low = Fn + min(Fi) + FIR + Fpe + min(Fp_th)
    G_low = Gn + min(Gi) + GIR + Gpe + GH2_val + min(Gp_th)
    xi_low = 8 * G_low / F_low**2 * tau_H_val * min(tau_ed_inv_val)

    F_high = Fn + max(Fi) + FIR + Fpe + max(Fp_th)
    G_high = Gn + max(Gi) + GIR + Gpe + GH2_val + max(Gp_th)
    xi_high = 8 * G_high / F_high**2 * tau_H_val * max(tau_ed_inv_val)

    omega_peak_low = omega_peak_th * np.sqrt(2 * G_low / F_low / (1 + np.sqrt(1 + xi_low)))
    omega_peak_high = omega_peak_th * np.sqrt(2 * G_high / F_high / (1 + np.sqrt(1 + xi_high)))

    # Array omega
    #aux_omega_min = 5e-3 * np.min((omega_peak_low, omega_peak_high))
    aux_omega_min = omega_min * (1+beta)
    aux_omega_max = 3 * np.max((omega_peak_low, omega_peak_high))
    aux_omega_max = np.min((aux_omega_max, omega_max * (1+beta)))
    aux_omega = makelogtab(aux_omega_min, aux_omega_max, Nomega) 
    Dln_omega = DX_over_X(aux_omega_min, aux_omega_max, Nomega) 

    # Fp(omega), Gp(omega)
    #FGp = FGp_averaged(env, a, beta, fZ, aux_omega, mu_ip, mu_op, tumbling=tumbling)
    #Fp = FGp['Fp'] # shape (Nomega, Nmu)
    #Gp = FGp['Gp']

    if use_spdust_plasma:
        FGp = spd_plasmadrag.FGp_averaged(env, a, fZ, aux_omega, mu_ip, mu_op, tumbling=tumbling)
        Fp = FGp['Fp'] # shape (Nomega, Nmu)
        Gp = FGp['Gp']
    else:
        FGp = FGp_averaged(env, a, beta, fZ, aux_omega, mu_ip, mu_op, tumbling=tumbling)
        Fp = FGp['Fp'] # shape (Nomega, Nmu)
        Gp = FGp['Gp']

    F = Fn + FIR + Fpe + np.matmul(np.ones((Nomega, 1)), Fi.reshape(1, -1)) + Fp
    G = Gn + GIR + Gpe + GH2_val + np.matmul(np.ones((Nomega, 1)), Gi.reshape(1, -1)) + Gp
    aux_result = aux_int(Tval, F, G, aux_omega, Nomega, Nmu, tau_H_val, tau_ed_inv_val, Inertia_val, Dln_omega) # shape: (Nmu, Nomega)
    log_Omega, aux_result = rescale_f_rot(aux_omega, aux_result, beta) 
    # Interpolate per mu
    result = np.zeros((Nmu, Nomega))
    omegaVec = makelogtab(omega_min, omega_max, Nomega)
    omegaVec_log = np.log(omegaVec)
    for ind in range(Nmu):
        aux_result_mu = aux_result[ind, :]
        interp_func = interp1d(log_Omega, aux_result_mu, kind='cubic', fill_value='extrapolate')
        result[ind, :] = interp_func(omegaVec_log)
    return result

    
def f_rot_old(env, a, beta, fZ, mu_ip, mu_op, tumbling=True, omega_min=1e7, omega_max=1e11, Nomega=1000):
    """
    Returns the rotational distribution function f_a:
    f(Omega | a, beta, mu)
    The function is calculated at frequencies centered on the approximate peak frequency
    of the emitted power. The output is a dictionary {omega, f_a} containing:
    - omega: array of frequencies.
    - f_a: rotational distribution function (2D array with dimensions [Nmu, Nomega]).

    Parameters:
    - env: Environmental parameters (T, nh).
    - a: Grain size.
    - ia: Index or identifier for intermediate value saving (optional).
    - fZ: Distribution of grain charges.
    - mu_ip: Electric dipole moment in-plane.
    - mu_op: Electric dipole moment out-of-plane.
    - tumbling: Boolean, whether tumbling disklike grains are considered.

    Returns:
    - Dictionary containing omega and f_a arrays.
    """

    if beta==0: # spherical grain
        tumbling = False

    Nmu = np.size(mu_ip)  # Number of dipole moments

    Inertia_val = Inertia_largest(a, beta)  # Grain's moment of inertia

    Tval, nH = env['T'], env['nh']

    # Characteristic timescales
    tau_H_val = tau_H(Tval, nH, a, beta)
    tau_ed_inv_val = tau_ed_inv(Tval, a, beta, mu_ip, mu_op, tumbling=tumbling)

    # Evaporation temperature
    Tev_val = Tev_effective(env, a)

    # F's and G's (except for plasma drag)
    FGn = FGn_averaged(env, a, beta, Tev_val, fZ, tumbling=tumbling)
    Fn = FGn['Fn']
    Gn = FGn['Gn']

    mu_tot = np.sqrt(mu_ip**2 + mu_op**2)
    FGi = FGi_averaged(env, a, beta, Tev_val, mu_tot, fZ)
    Fi = FGi['Fi']
    Gi = FGi['Gi']

    FGIR = FGIR_averaged(env, a, beta, fZ)
    FIR = FGIR['FIR']
    GIR = FGIR['GIR']

    FGpe = FGpe_averaged(env, a, beta, fZ)
    Fpe = FGpe['Fpe']
    Gpe = FGpe['Gpe']

    GH2_val = GH2(env, a, beta)

    omega = makelogtab(omega_min, omega_max, Nomega) * (1+beta)
    Dln_omega = DX_over_X(omega_min*(1+beta) , omega_max*(1+beta), Nomega) 

    # Fp(omega), Gp(omega)
    FGp = FGp_averaged(env, a, beta, fZ, omega, mu_ip, mu_op, tumbling=tumbling)
    Fp = FGp['Fp'] # shape (Nomega, Nmu)
    Gp = FGp['Gp']

    F = Fn + FIR + Fpe + np.matmul(np.ones((Nomega, 1)), Fi.reshape(1, -1)) + Fp
    G = Gn + GIR + Gpe + GH2_val + np.matmul(np.ones((Nomega, 1)), Gi.reshape(1, -1)) + Gp
    result1, result2 = aux_int(Tval, F, G, omega, Nomega, Nmu, tau_H_val, tau_ed_inv_val, Inertia_val, Dln_omega)
    return rescale_f_rot(result1, result2, beta)
   



# Normalized rotational damping and excitation rates through infrared emission.  
# Refers to Section 7 of Ali-Haimoud, Hirata & Dickinson, 2009, MNRAS, 395, 1055  
# and Section 6 of Silsbee, Ali-Haimoud & Hirata, 2010.      
# Use Draine & Li, 2001, ApJ 551, 807 (DL01) "thermal continuous" approximation 
# to compute the infrared spectrum.

import numpy as np
from . import SpDust_data_dir
from .util import maketab, DX_over_X, cgsconst, makelogtab
from .Grain import N_C, N_H, Inertia_largest, acx, grainparams
from .charge_dist import Qabs, nu_uisrf, Qabstabs
import os
from .mpiutil import *


class IR_arrays:
    a_min = None
    a_max = None
    Na =  None
    a_tab = None
    chi_min = None
    chi_max = None
    Nchi = None
    chi_tab = None
    FIR_integral_charged = None
    GIR_integral_charged = None
    FIR_integral_neutral = None
    GIR_integral_neutral = None
    Tev_tab = None

Theta_op = 863.0  # Debye temperature for out-of-plane C-C modes
Theta_ip = 2504.0  # Debye temperature for in-plane C-C modes
k = cgsconst.k
mp = cgsconst.mp
h = cgsconst.h
c = cgsconst.c
eV = cgsconst.eV
pi = np.pi
Energy_tab_max = np.max(Qabstabs.Qabs_hnu_tab) # Maximum energy (eV) in the Qabs_hnu_tab

#@njit
def f2(x):
    """
    Compute f2(x) as per DL01 Eq. (10). x can be an array.
    Parameters:
    - x: Input array.
    Returns:
    - Array of f2 values corresponding to the input x.
    """
    Nx = np.size(x)  # Number of elements in x
    Ny = 500     # Number of integration points for y
    y = maketab(0, 1, Ny)  # Create the y array from 0 to 1 with Ny points
    Dy = 1.0 / Ny  # Integration step size
    # Initialize the integrand array (Ny x Nx)
    integrand = np.zeros((Ny, Nx))
    # Create the y/x matrix (outer product of y and 1/x)
    y_over_x = np.outer(y, 1.0 / x)
    # Find the indices where y_over_x < 500
    ind = y_over_x < 500
    # Compute the integrand at the valid indices
    if np.any(ind):
       integrand[ind] = 1.0 / (np.exp(y_over_x[ind]) - 1.0)
    # Use loops for njit compatibility
    #for i in range(Ny):
    #    for j in range(Nx):
    #        if y_over_x[i, j] < 500:
    #            integrand[i, j] = 1.0 / (np.exp(y_over_x[i, j]) - 1.0)
    # Multiply by y^2
    integrand *= (y[:, np.newaxis]**2) 
    # Perform numerical integration: sum over the first axis (y-axis)
    result = 2.0 * Dy * np.sum(integrand, axis=0)
    return result

#@njit
def Energy_modes(a):
    Nc = N_C(a)
    Nm = (Nc - 2) * np.array([1, 2])  # Number of out-of-plane and in-plane modes
    beta2_tab = np.zeros(2)
    if Nc <= 54:
        beta2_tab = np.zeros(2)
    elif 54 < Nc <= 102:
        beta2_tab = (Nc - 54) / 52 / (2 * Nm - 1)
    else:
        beta2_tab = ((Nc - 2) / 52 * (102 / Nc)**(2 / 3) - 1) / (2 * Nm - 1)
    # Out-of-plane modes
    beta2 = beta2_tab[0]
    Nm = Nc - 2
    NewNm = Nm
    Eop_tab = np.zeros(NewNm + 1)
    delta_tab = 0.5 + np.zeros(NewNm + 1)
    delta_tab[2] = 1
    delta_tab[3] = 1
    Eop_tab[1:] = k * Theta_op * np.sqrt((1 - beta2) / Nm * (1 + np.arange(NewNm) - delta_tab[1:]) + beta2)
    # In-plane modes
    beta2 = beta2_tab[1]
    Nm = 2 * (Nc - 2)
    NewNm = Nm
    Eip_tab = np.zeros(NewNm + 1)
    delta_tab = 0.5 + np.zeros(NewNm + 1)
    delta_tab[2] = 1
    delta_tab[3] = 1
    Eip_tab[1:] = k * Theta_ip * np.sqrt((1 - beta2) / Nm * (1 + np.arange(NewNm) - delta_tab[1:]) + beta2)
    # C-H stretching modes
    lambda_inverse = np.array([886.0, 1161.0, 3030.0])
    CH_modes = h * c * lambda_inverse
    return Eop_tab[1:], Eip_tab[1:], CH_modes
    #return {'op': Eop_tab[1:], 'ip': Eip_tab[1:], 'CH': CH_modes}

#@njit
def EPAH(a, T, Energy_modes_op, Energy_modes_ip, Energy_modes_CH):
    """
    Compute the average energy E_PAH(a, T) in the thermal approximation.
    Parameters:
    - a: Grain size.
    - T: Array of temperatures. Must be ARRAY-like.
    - Energy_modes: Energy modes calculated based on grain size.
    Returns:
    - Array of average energies E_PAH for the given temperatures.
    """
    cutoff = 6e4  # Cutoff for Nc to switch between exact and continuous limit
    NT = np.size(T)  # Number of temperature values
    # Number of carbon and hydrogen atoms
    Nc = N_C(a)  # Assuming N_C(a) is defined
    NH = N_H(a)  # Assuming N_H(a) is defined
    # Energy of C-H bending and stretching modes
    E_Hmodes = np.zeros((np.size(Energy_modes_CH), NT))  # Initialize the energy array
    E_over_kT = Energy_modes_CH[:, None] / (k * T)  # Broadcasting to divide CH modes by k*T
    # Avoid floating point overflow for large E_over_kT values
    mask = E_over_kT < 600
    E_Hmodes[mask] = 1.0 / (np.exp(E_over_kT[mask]) - 1.0)
    # Use loops for njit compatibility
    """ 
    for i in range(E_over_kT.shape[0]):
        for j in range(E_over_kT.shape[1]):
            if E_over_kT[i, j] < 600:
                E_Hmodes[i, j] = 1.0 / (np.exp(E_over_kT[i, j]) - 1.0) 
    """
    # Multiply by CH modes to get total energy in these modes
    E_Hmodes = (Energy_modes_CH[:, None] * np.ones(NT)) * E_Hmodes
    E_Hmodes = NH * np.sum(E_Hmodes, axis=0)
    if Nc > cutoff:  # Continuous limit for large Nc, equation (33)
        return (Nc - 2) * k * (Theta_op * f2(T / Theta_op) + 2 * Theta_ip * f2(T / Theta_ip)) + E_Hmodes
    else:  # Exact mode spectrum for small Nc
        # Out-of-plane modes
        Eop_tab =  Energy_modes_op
        Eop_temp = np.zeros((np.size(Eop_tab), NT))
        E_over_kT = Eop_tab[:, None] / (k * T)
        mask = E_over_kT < 600
        Eop_temp[mask] = 1.0 / (np.exp(E_over_kT[mask]) - 1.0)
        # Use loops for njit compatibility
        """ 
        for i in range(E_over_kT.shape[0]):
            for j in range(E_over_kT.shape[1]):
                if E_over_kT[i, j] < 600:
                    Eop_temp[i, j] = 1.0 / (np.exp(E_over_kT[i, j]) - 1.0) 
        """
        Eop_temp = (Eop_tab[:, None] * np.ones(NT)) * Eop_temp
        Eop_bar = np.sum(Eop_temp, axis=0)
        # In-plane modes
        Eip_tab =  Energy_modes_ip
        Eip_temp = np.zeros((np.size(Eip_tab), NT))
        E_over_kT = Eip_tab[:, None] / (k * T)
        mask = E_over_kT < 600
        Eip_temp[mask] = 1.0 / (np.exp(E_over_kT[mask]) - 1.0)
        # Use loops for njit compatibility
        """ 
        for i in range(E_over_kT.shape[0]):
            for j in range(E_over_kT.shape[1]):
                if E_over_kT[i, j] < 600:
                    Eip_temp[i, j] = 1.0 / (np.exp(E_over_kT[i, j]) - 1.0)
        """
        Eip_temp = (Eip_tab[:, None] * np.ones(NT)) * Eip_temp
        Eip_bar = np.sum(Eip_temp, axis=0)
        return Eop_bar + Eip_bar + E_Hmodes
    
#@jit
def Temp(a, Energy, Energy_modes_op, Energy_modes_ip, Energy_modes_CH):
    Tmin = 1.0
    Tmax = 1.0e4
    NT = 100
    T_tab = makelogtab(Tmin, Tmax, NT)
    E_tab = EPAH(a, T_tab, Energy_modes_op, Energy_modes_ip, Energy_modes_CH)
    Temperature = np.exp(np.interp(np.log(Energy), np.log(E_tab), np.log(T_tab)))
    # Handle the energy modes (op, ip, CH)
    # modes = np.concatenate([ Energy_modes_op,  Energy_modes_ip, Energy_modes_CH])
    op_len = np.size(Energy_modes_op)
    ip_len = np.size(Energy_modes_ip)
    ch_len = np.size(Energy_modes_CH)
    modes = np.empty(op_len + ip_len + ch_len, dtype=np.float64)
    modes[:op_len] = Energy_modes_op
    modes[op_len:op_len+ip_len] = Energy_modes_ip
    modes[op_len+ip_len:] = Energy_modes_CH
    hbar_omega1 = np.min(modes)
    modes[np.argmin(modes)] = 2 * np.max(modes)
    for j in range(2, 21):
        hbar_omega20 = np.min(modes)
        modes[np.argmin(modes)] = 2 * np.max(modes)
    ind = np.where(Energy <= hbar_omega20)
    if len(ind[0]) > 0:
        Temperature[ind] = hbar_omega1 / (k * np.log(2.0))
    # Use loops for njit compatibility
    """ 
    for i in range(len(Energy)):
        if Energy[i] <= hbar_omega20:
            Temperature[i] = hbar_omega1 / (k * np.log(2.0)) 
    """
    return Temperature

#@njit
def Energy_bins(a, Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max):
    """
    Returns an array E_tab containing the central, minimum, and maximum values
    of M energy bins. It returns an array of shape (3, M+1) with the center, 
    minimum, and maximum of the energy bins.
    
    Parameters:
    - a: Grain size.
    - Energy_modes: Energy modes for the grain size.
    - M: Number of energy bins.
    - Energy_max: Maximum energy for the bins.
    
    Returns:
    - E_tab: A 2D NumPy array with the center, minimum, and maximum of energy bins.
    """
    jmax = 10  # First 10 energy bins are calculated as in DL01
    #assert M > 11, f'Must have M > 1. M = {M}.'
    # Collect modes 
    # modes = np.concatenate([ Energy_modes_op[:jmax+1],  Energy_modes_ip[:jmax+1], Energy_modes_CH])
    op_len = np.size(Energy_modes_op[:jmax+1])
    ip_len = np.size(Energy_modes_ip[:jmax+1])
    ch_len = np.size(Energy_modes_CH)
    # Preallocate the array for modes
    modes = np.empty(op_len + ip_len + ch_len, dtype=np.float64)
    # Fill the preallocated array
    modes[:op_len] = Energy_modes_op[:jmax+1]
    modes[op_len:op_len+ip_len] = Energy_modes_ip[:jmax+1]
    modes[op_len+ip_len:] = Energy_modes_CH
    hbar_omega = np.zeros(2 * jmax)
    # Fill hbar_omega with the minimum of the modes array
    for j in range(1, 2 * jmax):
        index = np.argmin(modes)
        hbar_omega[j] = modes[index]
        modes[index] = 2 * np.max(modes)
    # Initialize the E_tab array: (3, M+1) for center, min, and max values
    E_tab = np.zeros((3, M + 1))
    # Calculate the first few energy bins according to DL01
    E_tab[1, 1] = 1.5 * hbar_omega[1] - 0.5 * hbar_omega[2]
    for j in range(1, 3):
        E_tab[0, j] = hbar_omega[j]
        E_tab[1, j + 1] = 0.5 * (hbar_omega[j] + hbar_omega[j + 1])
        E_tab[2, j] = E_tab[1, j + 1]
    for j in range(3, jmax + 1):
        E_tab[2, j] = 0.5 * (hbar_omega[2 * j - 2] + hbar_omega[2 * j - 1])
        E_tab[0, j] = 0.5 * (hbar_omega[2 * j - 3] + hbar_omega[2 * j - 2])
        E_tab[1, j + 1] = E_tab[2, j]
    # Calculate the next bins (up to jmax) that are linearly spaced
    DeltaE = E_tab[2, jmax] - E_tab[1, jmax]
    # Define K_tab for linear spacing
    K_tab = 12 + np.arange(M - 11)
    # Calculate E_M_max for the highest energy bin
    E_M_max = np.exp((M - K_tab + 1) * np.log(E_tab[2, jmax] + (K_tab - 10) * DeltaE) - 
                     (M - K_tab) * np.log(E_tab[2, jmax] + (K_tab - 11) * DeltaE))
    # Check if the highest energy bin is below Energy_max
    if E_M_max[0] < Energy_max:
        #print(f'{M} energy bins are not enough, using {M+50} bins instead (in function Energy_bins).')
        return Energy_bins(a, Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M + 50, Energy_max)
    else:
        indices = np.where(E_M_max >= Energy_max)[0]  # Get indices where E_M_max >= Energy_max
        index = indices[-1]  # Find the largest index where E_M_max >= Energy_max
    BigK = K_tab[index]
    # Fill in the linearly spaced bins
    E_tab[2, jmax + 1:BigK + 1] = E_tab[2, jmax] + (np.arange(BigK + 1) - jmax)[jmax + 1:BigK + 1] * DeltaE
    E_tab[1, jmax + 2:BigK + 1] = E_tab[2, jmax + 1:BigK]
    E_tab[0, jmax + 1:BigK + 1] = 0.5 * (E_tab[1, jmax + 1:BigK + 1] + E_tab[2, jmax + 1:BigK + 1])
    # Fill in the last bins with logarithmic spacing if BigK < M
    if BigK < M:
        E_tab[2, BigK + 1:M + 1] = E_tab[2, BigK] * np.exp((np.arange(M + 1) - BigK)[BigK + 1:M + 1] * 
                                                           np.log(E_tab[2, BigK] / E_tab[2, BigK - 1]))
        E_tab[1, BigK + 1:M + 1] = E_tab[2, BigK:M]
        E_tab[0, BigK + 1:M + 1] = np.sqrt(E_tab[2, BigK + 1:M + 1] * E_tab[1, BigK + 1:M + 1])
    return E_tab

def Btilde_ISRF_func(a, Z, Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max):
    """
    Computes the matrix Bji for the average interstellar radiation field (ISRF),
    for a grain of radius `a` and charge `Z`, using the thermal continuous approximation.
    Parameters:
    - a: Grain radius.
    - Z: Grain charge.
    - Energy_modes: Dictionary containing energy modes.
    - M: Initial number of elements in energy bins.
    - Energy_max: Maximum energy.
    Returns:
    - Btilde matrix (upward and downward transition rates).
    """
    # Compute the energy bins
    Energies = Energy_bins(a, Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max)
    E_tab = Energies[0, :]  # Energy bins
    M = np.size(E_tab) - 1  # Update M based on energy bin calculation
    E_max = Energies[2, :]
    E_min = Energies[1, :]
    # Temperatures for the thermal approximation
    # T_tab = np.concatenate(([0], Temp(a, E_tab[1:],  Energy_modes_op, Energy_modes_ip, Energy_modes_CH)))
    aux_temp = Temp(a, E_tab[1:],  Energy_modes_op, Energy_modes_ip, Energy_modes_CH)
    T_tab = np.zeros(np.size(aux_temp)+1, dtype=np.float64)
    T_tab[1:] = aux_temp
    NEarr = 500  # Number of elements in the integral over energies
    Energy_min = h * c / 1e2  # Corresponds to a wavelength of 100 cm
    # Initialize transition rates arrays
    T_upward = np.zeros((M + 1, M + 1))  # Upward transition rates
    T_downward = np.zeros(M + 1)         # Downward transition rates
    T_downward_td = np.zeros((M + 1, M + 1))  # Thermal discrete downward rates
    # Pre-allocate arrays for the Cabs and nu_uisrf values
    Cabs = np.zeros(NEarr)
    E_array = np.zeros(NEarr)
    DE_over_E = np.zeros(NEarr)
    u_arr = np.zeros(NEarr)
    GuL = np.zeros(NEarr)
    # --- Upward Transitions ---
    for u in range(1, M + 1):
        Eu = E_tab[u]
        Emax_u = E_max[u]
        Emin_u = E_min[u]
        DeltaEu = Emax_u - Emin_u
        Tu = T_tab[u]
        # Case L = 0 (excitations from the ground state)
        W1 = Emin_u
        W4 = Emax_u
        E_array = makelogtab(W1, W4, NEarr)
        DE_over_E = DX_over_X(W1, W4, NEarr)
        Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
        T_upward[u, 0] = c / Eu * np.sum(nu_uisrf(E_array / eV) * Cabs) * DE_over_E
        T_downward_td[0, u] = 8 * pi / (h**3 * c**2) / Eu * (E_max[1] - E_min[1]) / DeltaEu * \
                              np.sum(E_array**4 / (np.exp(E_array / (k * Tu)) - 1) * Cabs) * DE_over_E
        # Case L > 0
        for L in range(1, u):
            EL = E_tab[L]
            DeltaEL = E_max[L] - E_min[L]
            W1 = Emin_u - E_max[L]
            W2 = min([Emin_u - E_min[L], Emax_u - E_max[L]])
            W3 = max([Emin_u - E_min[L], Emax_u - E_max[L]])
            W4 = Emax_u - E_min[L]
            if max([W1, Energy_min]) < W4:
                E_array = makelogtab(max([W1, Energy_min]), W4, NEarr)
                DE_over_E = DX_over_X(max([W1, Energy_min]), W4, NEarr)
                GuL = np.zeros(NEarr)
                ind1 = np.where((E_array > W1) & (E_array < W2))
                if len(ind1[0]) > 0:
                    GuL[ind1] = (E_array[ind1] - W1) / (DeltaEu * DeltaEL)
                ind2 = np.where((E_array > W2) & (E_array < W3))
                if len(ind2[0]) > 0:
                    GuL[ind2] = min([DeltaEu, DeltaEL]) / (DeltaEu * DeltaEL)
                ind3 = np.where((E_array > W3) & (E_array < W4))
                if len(ind3[0]) > 0:
                    GuL[ind3] = (W4 - E_array[ind3]) / (DeltaEu * DeltaEL)
                u_arr = nu_uisrf(E_array / eV)
                Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
                T_upward[u, L] = c * DeltaEu / (Eu - EL) * np.sum(GuL * Cabs * u_arr) * DE_over_E
                T_downward_td[L, u] = (8 * pi / (h**3 * c**2)) * DeltaEL / (Eu - EL) * \
                                      np.sum(GuL * Cabs * E_array**4 / (np.exp(E_array / (k * Tu)) - 1)) * DE_over_E
    # Case u = M, L > 0: include upward transitions above the highest level
    EM = E_tab[M]
    for Lind in range(1, M):
        EL = E_tab[Lind]
        W1 = E_min[M] - E_max[Lind]
        Wc = E_min[M] - E_min[Lind]
        integral_1 = 0
        if max([W1, Energy_min]) < Wc:
            E_array = makelogtab(max([W1, Energy_min]), Wc, NEarr)
            DE_over_E = DX_over_X(max([W1, Energy_min]), Wc, NEarr)
            u_arr = nu_uisrf(E_array / eV)
            Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
            integral_1 = np.sum((E_array - W1) / (Wc - W1) * Cabs * u_arr) * DE_over_E
        integral_2 = 0
        if Wc < 13.6 * eV:
            E_array = makelogtab(Wc, 13.6 * eV, NEarr)
            DE_over_E = DX_over_X(Wc, 13.6 * eV, NEarr)
            u_arr = nu_uisrf(E_array / eV)
            Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
            integral_2 = np.sum(Cabs * u_arr) * DE_over_E
        T_upward[M, Lind] = c / (EM - EL) * (integral_1 + integral_2)
    # Case u = M, L = 0
    Wc = E_min[M]
    integral_2 = 0
    if Wc < 13.6 * eV:
        E_array = makelogtab(Wc, 13.6 * eV, NEarr)
        DE_over_E = DX_over_X(Wc, 13.6 * eV, NEarr)
        u_arr = nu_uisrf(E_array / eV)
        Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
        integral_2 = np.sum(Cabs * u_arr) * DE_over_E
    T_upward[M, 0] = c / EM * integral_2
    # Intrabin contributions to upward transitions
    for u in range(2, M + 1):
        DeltaEu_1 = E_max[u - 1] - E_min[u - 1]
        if DeltaEu_1 > Energy_min:
            E_array = makelogtab(Energy_min, DeltaEu_1, NEarr)
            DE_over_E = DX_over_X(Energy_min, DeltaEu_1, NEarr)
            Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
            T_upward[u, u - 1] += c / (E_tab[u] - E_tab[u - 1]) * \
                                  np.sum((1 - E_array / DeltaEu_1) * Cabs * nu_uisrf(E_array / eV)) * DE_over_E
    # --- Downward Transitions ---
    for u in range(1, M + 1):
        T_downward[u] = 1 / (E_tab[u] - E_tab[u - 1]) * np.sum((E_tab[u] - E_tab[:u]) * T_downward_td[:u, u])
        DeltaEu = E_max[u] - E_min[u]
        if Energy_min < DeltaEu:
            E_array = makelogtab(Energy_min, DeltaEu, NEarr)
            DE_over_E = DX_over_X(Energy_min, DeltaEu, NEarr)
            Cabs = pi * a**2 * Qabs(a, Z, E_array / eV)
            T_downward[u] += 8 * pi / (h**3 * c**2) / (E_tab[u] - E_tab[u - 1]) * \
                             np.sum((1 - E_array / DeltaEu) * E_array**4 * Cabs / (np.exp(E_array / (k * T_tab[u])) - 1)) * DE_over_E
    # Define Btilde_{j,i}
    Tinv = np.flip(T_upward, axis=0)
    Binv = np.cumsum(Tinv, axis=0)
    Btab = np.flip(Binv, axis=0)
    T_downward = np.outer(T_downward, np.ones(M + 1))
    T_downward[0, :] = np.ones(M + 1)
    Btilde = Btab / T_downward
    return Btilde, M

def distribution(a, Z, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max):
    """
    Returns the distribution function for the energy levels.
    Parameters:
    - a: Grain radius.
    - Z: Grain charge.
    - Chi: ISRF intensity scaling factor.
    -  Energy_modes_op, Energy_modes_ip, Energy_modes_CH: Energy modes calculated based on grain size.
    - M: Initial number of elements in energy bins.
    - Energy_max: Maximum energy.
    Returns:
    - Dictionary containing Energy_bins and P_tab (the distribution).
    """
    # Iterate while X_tab[M] is greater than a small threshold (1e-14)
    # Scale the initial Energy_max
    Energy_max = Energy_max/3
    # Initialize the distribution (X_tab) with ones
    X_tab = np.ones(M + 1)
    while X_tab[M] > 1e-14:    
        Energy_max *= 3
        aux1, M = Btilde_ISRF_func(a, Z,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max) 
        if rank0:
            print(f"E_M = {Energy_max / eV:.2f} eV is not high enough in 'distribution'. Using {3 * Energy_max / eV:.2f} eV instead")
        Btilde = Chi * aux1 
        X_tab = np.ones(M + 1)  # Reset X_tab since M might have changed
        
        # Loop through energy levels and compute the distribution
        for j in range(1, M + 1):
            X_tab[j] = np.sum(Btilde[j, 0:j] * X_tab[0:j])  # Reform array and compute sum
            X_tab[0:j + 1] =  X_tab[0:j + 1] / np.sum(X_tab[0:j + 1])  # Normalize the distribution
    
    '''
    Energy_max = Energy_tab_max
    X_tab = np.ones(M + 1)  # Initialize the distribution with ones
    while X_tab[M] > 1e-14: 
        M+=50
        aux1, M = Btilde_ISRF_func(a, Z,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max) 
        X_tab = np.ones(M + 1)
        Btilde = Chi * aux1 
        for j in range(1, M + 1):
            X_tab[j] = np.sum(Btilde[j, 0:j] * X_tab[0:j])  # Reform array and compute sum
            X_tab[0:j + 1] =  X_tab[0:j + 1] / np.sum(X_tab[0:j + 1])  # Normalize the distribution
    
    
    aux1, M = Btilde_ISRF_func(a, Z,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max) 
    X_tab = np.ones(M + 1)
    Btilde = Chi * aux1 
    for j in range(1, M + 1):
        X_tab[j] = np.sum(Btilde[j, 0:j] * X_tab[0:j])  # Reform array and compute sum
        X_tab[0:j + 1] =  X_tab[0:j + 1] / np.sum(X_tab[0:j + 1])  # Normalize the distribution
    '''
    # Get the energy bins
    Energy_bins_eval = Energy_bins(a,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, M, Energy_max)  
    # Return the energy bins and the computed distribution (P_tab)
    return {'Energy_bins': Energy_bins_eval, 'P_tab': X_tab}, Energy_max, M

def IRemission(a, Z, nu_tab,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, distribution):
    """
    Computes the infrared emission spectrum for a grain of radius `a` and charge `Z`.
    Parameters:
    - a: Grain radius.
    - Z: Grain charge.
    - Chi: ISRF intensity scaling factor.
    - nu_tab: Array of frequency values.
    -  Energy_modes_op, Energy_modes_ip, Energy_modes_CH: Energy modes calculated based on grain size.
    - distribution: Distribution function calculated for energy levels.
    Returns:
    - Array of infrared emission values (F_nu) corresponding to the given frequencies (nu_tab).
    """
    # Extract the distribution data
    P_tab = distribution['P_tab']
    Mval = np.size(P_tab) - 1
    Energy_bins_val = distribution['Energy_bins']
    E_tab = Energy_bins_val[0, 1:Mval+1].reshape(Mval)  # Reforming as needed for Python
    T_tab = Temp(a, E_tab,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH)  
    N_nu = np.size(nu_tab)
    F_nu = np.zeros(N_nu)  # Initialize the emission array
    # Broadcasting arrays for matrix multiplication
    new_nu_tab = nu_tab[:, None] * np.ones((1, Mval))
    new_E_tab = np.ones((N_nu, 1)) * E_tab[None, :]
    new_P_tab = np.ones((N_nu, 1)) * P_tab[1:Mval+1][None, :]
    new_T_tab = np.ones((N_nu, 1)) * T_tab[None, :]
    # Handle cases to avoid floating point overflow
    ind = np.where((new_E_tab < h * new_nu_tab) | (h * new_nu_tab / (k * new_T_tab) > 700))
    if ind[0].size > 0:
        new_P_tab[ind] = 0  # Set P_tab values to zero where overflow occurs
        new_nu_tab[ind] = 700 * k * new_T_tab[ind] / h  # Adjust nu_tab to avoid overflow
        F_nu = np.sum(new_P_tab / (np.exp(h * new_nu_tab / (k * new_T_tab)) - 1), axis=1)
    # Compute the final emission spectrum
    F_nu = (2 * h * nu_tab**3 / c**2) * pi * a**2 * Qabs(a, Z, h * nu_tab / eV) * F_nu
    return F_nu

def FGIR_integrals(a, Z, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, Mval, Energy_max):
    """
    Computes the integrals \int F_nu/nu^2 dnu and \int F_nu/nu dnu.
    Parameters:
    - a: Grain radius.
    - Z: Grain charge.
    - Chi: ISRF intensity scaling factor.
    -  Energy_modes_op, Energy_modes_ip, Energy_modes_CH: Energy modes calculated based on grain size.
    - M: Initial number of elements in energy bins.
    - Energy_max: Maximum energy.
    Returns:
    - A dictionary containing the FIR and GIR integrals.
    """
    # Compute the distribution
    dist, Energy_max, Mval = distribution(a, Z, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, Mval, Energy_max)
    # Update M and Energy_max based on the distribution function
    lambda_max = 1e-1  # 0.1 cm
    nu_min = c / lambda_max
    nu_max = Energy_max / h
    N_lambda = 1000  # Number of frequency points
    # Generate the logarithmic frequency table and differential factor
    nu_tab = makelogtab(nu_min, nu_max, N_lambda)
    Dnu_over_nu = DX_over_X(nu_min, nu_max, N_lambda)  
    # Compute the infrared emission spectrum
    F_nu = IRemission(a, Z, nu_tab,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, dist)
    # Perform the integrals
    FIR_integral = np.sum(F_nu / nu_tab) * Dnu_over_nu
    GIR_integral = np.sum(F_nu) * Dnu_over_nu
    return {'FIR_integral': FIR_integral, 'GIR_integral': GIR_integral}, Energy_max, Mval

# Function to set up the infrared arrays
def set_up_IR_arrays():
    """
    Sets up the infrared arrays for the calculations.
    
    Returns:
    - Dictionary containing the parameters and the arrays.
    """
    # Grain size range
    a_min = 3.5e-8  # cm
    a_max = 1e-6    # cm
    Na = 30
    a_tab = makelogtab(a_min, a_max, Na)
    # Scaling factor (chi) range
    chi_0 = 1e-5
    chi_N = 1e10
    Nchi = 30
    # Calculating chi_min and chi_max as done in IDL
    chi_min = np.exp(np.log(chi_0) - 1.0 / (2 * Nchi) * np.log(chi_N / chi_0))
    chi_max = np.exp(np.log(chi_N) - 1.0 / (2 * Nchi) * np.log(chi_N / chi_0))
    # Generating the chi_tab array
    chi_tab = makelogtab(chi_min, chi_max, Nchi)
    IR_arrays.a_min = a_min
    IR_arrays.a_max = a_max
    IR_arrays.Na = Na
    IR_arrays.a_tab = a_tab
    IR_arrays.chi_min = chi_min
    IR_arrays.chi_max = chi_max
    IR_arrays.Nchi = Nchi
    IR_arrays.chi_tab = chi_tab
    pass

set_up_IR_arrays()

def compute_FGIR_integrals():
    """
    Computes and stores FIR and GIR integrals for charged and neutral grains
    over a range of grain sizes (a) and ISRF scaling factors (Chi).
    """
    Na = IR_arrays.Na
    Nchi = IR_arrays.Nchi
    a_tab = IR_arrays.a_tab
    chi_tab = IR_arrays.chi_tab
    # If the FIR and GIR integrals are already stored in the data directory, 
    # load them and return
    bool1 = os.path.exists(SpDust_data_dir + f'FIR_integral_charged_{Na}a_{Nchi}chi.txt')
    bool2 = os.path.exists(SpDust_data_dir + f'GIR_integral_charged_{Na}a_{Nchi}chi.txt')
    bool3 = os.path.exists(SpDust_data_dir + f'FIR_integral_neutral_{Na}a_{Nchi}chi.txt')
    bool4 = os.path.exists(SpDust_data_dir + f'GIR_integral_neutral_{Na}a_{Nchi}chi.txt')
    if bool1 and bool2 and bool3 and bool4:
        FIR_integral_charged = np.loadtxt(SpDust_data_dir + f'FIR_integral_charged_{Na}a_{Nchi}chi.txt')
        GIR_integral_charged = np.loadtxt(SpDust_data_dir + f'GIR_integral_charged_{Na}a_{Nchi}chi.txt')
        FIR_integral_neutral = np.loadtxt(SpDust_data_dir + f'FIR_integral_neutral_{Na}a_{Nchi}chi.txt')
        GIR_integral_neutral = np.loadtxt(SpDust_data_dir + f'GIR_integral_neutral_{Na}a_{Nchi}chi.txt')
        IR_arrays.FIR_integral_charged = FIR_integral_charged
        IR_arrays.GIR_integral_charged = GIR_integral_charged
        IR_arrays.FIR_integral_neutral = FIR_integral_neutral
        IR_arrays.GIR_integral_neutral = GIR_integral_neutral
        return
    # Loop over grain sizes (a_tab) and Chi values (chi_tab)
    # Define a function to do the loop for a given grain size
    def loop(ia):
        a = a_tab[ia]
        Energy_modes_op, Energy_modes_ip, Energy_modes_CH = Energy_modes(a)
        Mval = 100
        Energy_max = 13.6 * eV  # Maximum energy in erg
        result = np.zeros((Nchi, 4))
        for ichi in range(Nchi):
            Chi = chi_tab[ichi]
            # Compute integrals for charged grains (Z = 1)
            FGIR_integral_charged, Energy_max, Mval = FGIR_integrals(a, 1, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, Mval, Energy_max)
            result[ichi, 0] = FGIR_integral_charged['FIR_integral']
            result[ichi, 1] = FGIR_integral_charged['GIR_integral']
            # Compute integrals for neutral grains (Z = 0)
            FGIR_integral_neutral, Energy_max, Mval = FGIR_integrals(a, 0, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, Mval, Energy_max)
            result[ichi, 2] = FGIR_integral_neutral['FIR_integral']
            result[ichi, 3] = FGIR_integral_neutral['GIR_integral']
        return result
    
    # Parallelize the loop
    result = np.array(parallel_map(loop, np.arange(Na)) ) # shape: (Na, Nchi, 4)
        
    """
    # Initialize arrays for FIR and GIR integrals (charged and neutral)
    FIR_integral_charged = np.zeros((Na, Nchi))
    GIR_integral_charged = np.zeros((Na, Nchi))
    FIR_integral_neutral = np.zeros((Na, Nchi))
    GIR_integral_neutral = np.zeros((Na, Nchi))
    for ia in range(Na):
        a = a_tab[ia]
        Energy_modes_op, Energy_modes_ip, Energy_modes_CH = Energy_modes(a)  
        Mval = 100
        Energy_max = 13.6 * eV  # Maximum energy in erg
        for ichi in range(Nchi):
            Chi = chi_tab[ichi]
            # Compute integrals for charged grains (Z = 1)
            FGIR_integral_charged, Energy_max, Mval = FGIR_integrals(a, 1, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, Mval, Energy_max)
            FIR_integral_charged[ia, ichi] = FGIR_integral_charged['FIR_integral']
            GIR_integral_charged[ia, ichi] = FGIR_integral_charged['GIR_integral']
            # Compute integrals for neutral grains (Z = 0)
            FGIR_integral_neutral, Energy_max, Mval = FGIR_integrals(a, 0, Chi,  Energy_modes_op, Energy_modes_ip, Energy_modes_CH, Mval, Energy_max)
            FIR_integral_neutral[ia, ichi] = FGIR_integral_neutral['FIR_integral']
            GIR_integral_neutral[ia, ichi] = FGIR_integral_neutral['GIR_integral']
    IR_arrays.FIR_integral_charged = FIR_integral_charged
    IR_arrays.GIR_integral_charged = GIR_integral_charged
    IR_arrays.FIR_integral_neutral = FIR_integral_neutral
    IR_arrays.GIR_integral_neutral = GIR_integral_neutral
    """
    if rank0:
        IR_arrays.FIR_integral_charged = result[:, :, 0]
        IR_arrays.GIR_integral_charged = result[:, :, 1]
        IR_arrays.FIR_integral_neutral = result[:, :, 2]
        IR_arrays.GIR_integral_neutral = result[:, :, 3]
        # Save FIR and GIR integrals for charged grains
        np.savetxt(SpDust_data_dir + f'FIR_integral_charged_{Na}a_{Nchi}chi.txt', IR_arrays.FIR_integral_charged)
        np.savetxt(SpDust_data_dir + f'GIR_integral_charged_{Na}a_{Nchi}chi.txt', IR_arrays.GIR_integral_charged)
        # Save FIR and GIR integrals for neutral grains
        np.savetxt(SpDust_data_dir + f'FIR_integral_neutral_{Na}a_{Nchi}chi.txt', IR_arrays.FIR_integral_neutral)
        np.savetxt(SpDust_data_dir + f'GIR_integral_neutral_{Na}a_{Nchi}chi.txt', IR_arrays.GIR_integral_neutral)
    return

"""
def precompile_njit_functions():
    a=3e-7
    op, ip, ch = Energy_modes(a)
    aux = Energy_bins(a, op, ip, ch, 100, 13.6*eV)
    aux = Temp(a,np.array([50]), op, ip, ch)
    pass
precompile_njit_functions()
"""

compute_FGIR_integrals()

def FGIR_integrals_interpol(a, Z, Chi):
    """
    Interpolates or extrapolates the FIR and GIR integrals for a given grain size `a`, charge `Z`, and ISRF scaling factor `Chi`.
    Parameters:
    - a: Grain size.
    - Z: Grain charge.
    - Chi: ISRF intensity scaling factor.
    
    Returns:
    - A dictionary containing the interpolated/extrapolated FIR and GIR integrals.
    """
    FIR_integral_charged, GIR_integral_charged, FIR_integral_neutral, GIR_integral_neutral, \
    a_min, a_max, Na, chi_min, chi_max, Nchi \
    = IR_arrays.FIR_integral_charged, IR_arrays.GIR_integral_charged, IR_arrays.FIR_integral_neutral, IR_arrays.GIR_integral_neutral, \
        IR_arrays.a_min, IR_arrays.a_max, IR_arrays.Na, IR_arrays.chi_min, IR_arrays.chi_max, IR_arrays.Nchi
    
    # Select the appropriate FIR and GIR integral tables based on charge
    if Z != 0:
        FIR_integral_tab = FIR_integral_charged
        GIR_integral_tab = GIR_integral_charged
    else:
        FIR_integral_tab = FIR_integral_neutral
        GIR_integral_tab = GIR_integral_neutral
    # Tabulation for grain size (a_tab) and chi (chi_tab)
    a_tab = IR_arrays.a_tab
    chi_tab = IR_arrays.chi_tab

    # Finding the indices and coefficients alpha and beta for a and chi s.t. a = alpha *a_i + (1-alpha)*a_i+1 
    if a <= min(a_tab):
        ia = 0
        alpha = 1.0
    elif a >= max(a_tab):
        ia = Na - 2
        alpha = 0.0
    else:
        ia = np.where(a_tab <= a)[0][-1]
        alpha = 1.0 - Na * np.log(a / a_tab[ia]) / np.log(a_max / a_min)

    # --- Case of low radiation field: FIR, GIR are linear in chi ---
    if Chi <= np.min(chi_tab):
        FIR_integral_0 = np.exp(alpha * np.log(FIR_integral_tab[ia, 0]) + (1.0 - alpha) * np.log(FIR_integral_tab[ia+1, 0]))
        FIR_integral = (Chi / chi_tab[0]) * FIR_integral_0
        GIR_integral_0 = np.exp(alpha * np.log(GIR_integral_tab[ia, 0]) + (1.0 - alpha) * np.log(GIR_integral_tab[ia+1, 0]))
        GIR_integral = (Chi / chi_tab[0]) * GIR_integral_0
        return {'FIR_integral': FIR_integral, 'GIR_integral': GIR_integral}
    
    # --- Case of high radiation field: Approximate FIR, GIR by a power-law ---
    if Chi >= np.max(chi_tab):
        if rank0:
            print(f'You are using chi= {Chi}! The code is written for chi < 1E5. It assumes a power-law in chi for higher values.')
        FIR_integral_last = np.exp(alpha * np.log(FIR_integral_tab[ia, Nchi-1]) 
                                   + (1.0 - alpha) * np.log(FIR_integral_tab[ia+1, Nchi-1]))
        FIR_integral_before_last = np.exp(alpha * np.log(FIR_integral_tab[ia, Nchi-2]) 
                                          + (1.0 - alpha) * np.log(FIR_integral_tab[ia+1, Nchi-2]))
        
        FIR_index = np.log(FIR_integral_last / FIR_integral_before_last) / np.log(chi_tab[Nchi-1] / chi_tab[Nchi-2])
        FIR_integral = (Chi / chi_tab[Nchi-1]) ** FIR_index * FIR_integral_last
        GIR_integral_last = np.exp(alpha * np.log(GIR_integral_tab[ia, Nchi-1]) 
                                   + (1.0 - alpha) * np.log(GIR_integral_tab[ia+1, Nchi-1]))
        GIR_integral_before_last = np.exp(alpha * np.log(GIR_integral_tab[ia, Nchi-2]) 
                                          + (1.0 - alpha) * np.log(GIR_integral_tab[ia+1, Nchi-2]))
        
        GIR_index = np.log(GIR_integral_last / GIR_integral_before_last) / np.log(chi_tab[Nchi-1] / chi_tab[Nchi-2])
        GIR_integral = (Chi / chi_tab[Nchi-1]) ** GIR_index * GIR_integral_last
        return {'FIR_integral': FIR_integral, 'GIR_integral': GIR_integral}
    
    # Case for intermediate Chi values (interpolation)
    ichi = np.where(chi_tab <= Chi)[0][-1]
    beta = 1.0 - Nchi * np.log(Chi / chi_tab[ichi]) / np.log(chi_max / chi_min)
    FIR_integral = np.exp(
        alpha * (beta * np.log(FIR_integral_tab[ia, ichi]) 
                 + (1 - beta) * np.log(FIR_integral_tab[ia, ichi + 1])) 
                 + (1 - alpha) * (beta * np.log(FIR_integral_tab[ia + 1, ichi]) 
                                  + (1 - beta) * np.log(FIR_integral_tab[ia + 1, ichi + 1]))
    )
    GIR_integral = np.exp(
        alpha * (beta * np.log(GIR_integral_tab[ia, ichi]) 
                 + (1 - beta) * np.log(GIR_integral_tab[ia, ichi + 1])) 
                 + (1 - alpha) * (beta * np.log(GIR_integral_tab[ia + 1, ichi]) 
                                  + (1 - beta) * np.log(GIR_integral_tab[ia + 1, ichi + 1]))
    )
    return {'FIR_integral': FIR_integral, 'GIR_integral': GIR_integral}

def FGIR(env, a, beta, Zg):
    """
    Computes FIR and GIR for a grain with radius `a`, charge `Zg`, and environment `env`.
    Parameters:
    - env: Environment object containing Chi, T (temperature), and nh (number density of hydrogen).
    - a: Grain radius.
    - Zg: Grain charge.
    Returns:
    - A dictionary containing FIR and GIR values.
    """
    # Extract environment parameters
    Chi = env['Chi']
    Tval = env['T']
    nh = env['nh']
    # Compute various physical quantities
    Inertia_val = Inertia_largest(a, beta) 
    acx_val = acx(a, beta)  
    tau_H = 1.0 / (nh * mp * np.sqrt(2.0 * k * Tval / (pi * mp)) 
                   * 4.0 * pi * acx_val**4 / (3.0 * Inertia_val))
    # Interpolate FIR and GIR integrals
    FGIR_integrals_aux = FGIR_integrals_interpol(a, Zg, Chi)  
    IntF = FGIR_integrals_aux['FIR_integral']
    IntG = FGIR_integrals_aux['GIR_integral']
    # Compute FIR and GIR values
    FIR = 2.0 * tau_H / (pi * Inertia_val) * IntF
    GIR = h * tau_H / (3.0 * pi * Inertia_val * k * Tval) * IntG  # Corrected version
    # Adjust for disklike grains if a < a2
    if a < grainparams.a2:
        FIR *= 5.0 / 3.0
    return {'FIR': FIR, 'GIR': GIR}

def FGIR_averaged(env, a, beta, fZ):
    """
    Computes the averaged FIR and GIR values for a grain with radius `a`, using the charge distribution `fZ` and environment `env`.
    Parameters:
    - env: Environment object containing Chi, T (temperature), and nh (number density of hydrogen).
    - a: Grain radius.
    - fZ: Charge distribution array.
    Returns:
    - A dictionary containing the averaged FIR and GIR values.
    """
    # Extract the first element of fZ (f0 corresponds to the neutral grain contribution)
    f0 = fZ[1, 0]
    # Compute FIR and GIR for neutral and charged grains
    FGIR_neutral = FGIR(env, a, beta, 0)  # For neutral grain (Z = 0)
    FGIR_charged = FGIR(env, a, beta, 1)  # For charged grain (Z = 1)
    # Average the FIR and GIR values
    FIR = f0 * FGIR_neutral['FIR'] + (1.0 - f0) * FGIR_charged['FIR']
    GIR = f0 * FGIR_neutral['GIR'] + (1.0 - f0) * FGIR_charged['GIR']
    return {'FIR': FIR, 'GIR': GIR}


from .util import cgsconst, DX_over_X, maketab, makelogtab 
from .mpiutil import *
from .Grain import grainparams, rms_dipole
from .charge_dist import charge_dist
from .AngMomDist import log_f_rot
import numpy as np
from scipy.interpolate import interp1d


pi = np.pi
c = cgsconst.c
h = cgsconst.h
k = cgsconst.k
eV = cgsconst.eV
me = cgsconst.me
q = cgsconst.q
mp = cgsconst.mp

def mu2_f_cond(env, a, beta, fZ, mu_rms, ip, Ndipole, tumbling=True, omega_min=1e7, omega_max=1e12, Nomega=1000, spdust_plasma=False):
    """
    Returns a [3, Nomega] array:
    [omega, <mu_ip^2 fa(omega)>, <mu_op^2 fa(omega)>].
    
    Parameters:
    - env: Environment parameters (such as temperature).
    - a: Grain size.
    - beta: Grain shape parameter.
    - fZ: Distribution of grain charges.
    - mu_rms: Root mean square dipole moment.
    - ip: Fraction of in-plane dipole moment.
    - Ndipole: Number of dipole values used in averaging.
    - tumbling: Boolean to indicate whether the grains are tumbling disklike grains.
    Returns:
    - A numpy array [2, Nomega] containing omega, <mu_ip^2 fa(omega)>, and <mu_op^2 fa(omega)>.
    """
    op = 1.0 - ip  # Out-of-plane dipole moment fraction

    if Ndipole == 1:  # Not averaging over dipoles
        mu_ip = np.array([np.sqrt(ip) * mu_rms]) # shape (1,)
        mu_op = np.array([np.sqrt(op) * mu_rms]) # shape (1,)
        log_f_a = log_f_rot(env, a, beta, fZ, mu_ip, mu_op, 
                           tumbling=tumbling, 
                           omega_min=omega_min, omega_max=omega_max, Nomega=Nomega, use_spdust_plasma=spdust_plasma)
        f_a = np.exp(log_f_a)
        mu_ip2_fa = f_a * mu_ip[0]**2
        mu_op2_fa = f_a * mu_op[0]**2
    else:
        # Set up for averaging
        xmin = 5e-3
        xmed = 0.5
        xmax = 5.0
        aux_Nd = int(Ndipole/2)
        x_tab = np.concatenate((makelogtab(xmin, xmed, aux_Nd), 
                                maketab(xmed, xmax, aux_Nd)))
        Dx_tab = np.concatenate((DX_over_X(xmin, xmed, aux_Nd) * makelogtab(xmin, xmed, aux_Nd), 
                                 (xmax - xmed)/aux_Nd + np.zeros(aux_Nd)))
        if a < grainparams.a2:  # Disk-like grains, need 2D Gaussian
            mu_ip = np.sqrt(ip) * mu_rms * np.outer(x_tab, np.ones(Ndipole))
            mu_op = np.sqrt(op) * mu_rms * np.outer(np.ones(Ndipole), x_tab)
            Dmu_ip = np.outer(Dx_tab, np.ones(Ndipole))
            Dmu_op = Dmu_ip.T
            # Probability calculation for dipoles
            if ip == 0.0:
                Proba = np.exp(-0.5 * mu_op**2 / mu_rms**2) * Dmu_op
            elif op == 0.0:
                Proba = mu_ip / mu_rms * np.exp(-mu_ip**2 / mu_rms**2) * Dmu_ip
            else:
                Proba = mu_ip / mu_rms * np.exp(-mu_ip**2 / (ip * mu_rms**2)) * \
                         np.exp(-0.5 * mu_op**2 / (op * mu_rms**2)) * Dmu_ip * Dmu_op
            Proba = Proba / np.sum(Proba)
            # Flatten 2D arrays to 1D
            mu_ip = mu_ip.flatten()
            mu_op = mu_op.flatten()
            Proba = Proba.flatten()
            log_f_a = log_f_rot(env, a, beta, fZ, mu_ip, mu_op, 
                               tumbling=tumbling,  
                               omega_min=omega_min, 
                               omega_max=omega_max, 
                               Nomega=Nomega,
                               use_spdust_plasma=spdust_plasma)
            f_a = np.exp(log_f_a)
            Proba = np.outer(Proba, np.ones(Nomega))
            mu_ip = np.outer(mu_ip, np.ones(Nomega)) # shape (Ndipole, Nomega)
            mu_op = np.outer(mu_op, np.ones(Nomega))
            # Calculate <mu_ip^2 fa(omega)> and <mu_op^2 fa(omega)>
            mu_ip2_fa = np.sum(mu_ip**2  * Proba * f_a, axis=0) # sum over Ndipole
            mu_op2_fa = np.sum(mu_op**2  * Proba * f_a, axis=0)
            #mu_ip2_fa = np.sum((mu_ip**2 / mu_rms**2) * Proba * f_a, axis=0) # sum over Ndipole
            #mu_op2_fa = np.sum((mu_op**2 / mu_rms**2) * Proba * f_a, axis=0)
            #if ip != 0.0 and op != 0.0:
            #    mu_ip2_fa /= ip
            #    mu_op2_fa /= op
        else:  # Spherical grains, average over grain orientation first
            Proba = x_tab**2 * np.exp(-1.5 * x_tab**2) * Dx_tab
            Proba = Proba / np.sum(Proba)
            log_f_a = log_f_rot(env, a, beta, fZ, np.sqrt(2/3) * mu_rms * x_tab, mu_rms / np.sqrt(3) * x_tab, 
                               tumbling=tumbling,  omega_min=omega_min, omega_max=omega_max, Nomega=Nomega, use_spdust_plasma=spdust_plasma)
            f_a = np.exp(log_f_a)
            Proba = np.outer(Proba, np.ones(Nomega))
            x_tab = np.outer(x_tab, np.ones(Nomega))
            mu2_fa_aux = np.sum(x_tab**2 * Proba * f_a, axis=0)
            mu_ip2_fa = mu2_fa_aux * mu_rms**2 * 2/3
            mu_op2_fa = mu2_fa_aux * mu_rms**2 * 1/3
    # Prepare the result as a [3, Nomega] array
    result = np.zeros((2, Nomega))
    
    result[0, :] = mu_ip2_fa
    result[1, :] = mu_op2_fa
    # Save the intermediate results (optional)
    # print('Saving results...')
    # np.savez(f'{ia}_mu_fa.npz', omega=omega, mu_ip2_fa=mu_ip2_fa, mu_op2_fa=mu_op2_fa, mu_rms=mu_rms, ip=ip)
    return result

def mu2_f(env, a_tab, beta_tab, f_a_beta, mole_dipole, ip, Ndipole, tumbling=True, parallel=True, contract_a=True, omega_min=1e7, omega_max=1e15, Nomega=1000, spdust_plasma=False):
    '''
    Calculate mu^2 f(omega, a, beta, mu) and marginalize over "mu" and "a" (if contract_a is True).
    '''
    num_a = np.size(a_tab)
    num_beta = np.size(beta_tab)
    
    def aux_func(a_ind):
        result = []
        a = a_tab[a_ind]
        for beta_ind in range(num_beta):
            beta = beta_tab[beta_ind]
            if f_a_beta[a_ind, beta_ind] == 0.:
                result.append(np.zeros((2, Nomega)))
                continue
            fZ = charge_dist(env, a, beta)
            Z2 = np.sum(fZ[0, :]**2 * fZ[1, :])
            mu_rms = rms_dipole(a, beta, Z2, mole_dipole)
            mu2_fa_aux  = mu2_f_cond(env, a, beta, fZ, mu_rms, ip, Ndipole, tumbling=tumbling, omega_min=omega_min, omega_max=omega_max, Nomega=Nomega, spdust_plasma=spdust_plasma)
            result.append(mu2_fa_aux)
        return np.array(result)
    
    if parallel:
        aux = np.array(parallel_map(aux_func, np.arange(num_a)))
    else:
        aux=[]
        for ai in np.arange(num_a):
            aux.append(aux_func(ai))
        aux = np.array(aux)
    result = aux * f_a_beta[:,:,np.newaxis,np.newaxis]
    if contract_a:
        # Only when f_a_beta incorporate the differential element da 
        # we can simply sum over a as the integral over a.
        # This is the case when the keyword "a_weighted" is set to True in the function size_and_shape_dist. (See main.Grain)
        return np.sum(result, axis=0) # shape (Nbeta, 2, Nomega)
    else:
        return result # shape (Na, Nbeta, 2, Nomega)
    
def SED(nu_tab, mu2_f_arr, beta_tab, angular_Omega_tab, cos_theta_list, cos_theta_weights):
    '''
    Calculate the SED for an ensemble of spinning dust grains.
    Parameters:
    - nu_tab: array of frequencies in Hz.
    - mu2_f_arr: array of shape (Nbeta, 2, Nomega), 
                where (:, 0, :) and (:, 1, :) are mu2_f_ip and mu2_f_op, respectively:
                    - mu2_f_ip := mu_perp^2 f(omega, a, beta, mu) marginalized over grain sizes (a) and dipole moments (mu) distributions
                    - mu2_f_op := mu_parallel^2 f(omega, a, beta, mu) marginalized over grain sizes (a) and dipole moments (mu) distributions
    - beta_tab: array of grain shape parameters, shape (Nbeta,)
    - angular_Omega_tab: array of rescaled angular momentum, shape (Nomega,). 
                This corresponds to the last axis of mu2_f_arr.
    - cos_theta_list: array of cos(theta) values for averaging over nutation angle, shape (N_cos_theta,)
    - cos_theta_weights: array of weights for averaging over nutation angle, shape (Nbeta, N_cos_theta)
                Note the shape: we allow for different weights for each beta value.
    '''
    mu2_f_ip = mu2_f_arr[:, 0, :]
    mu2_f_op = mu2_f_arr[:, 1, :]
    n_cos_theta = np.size(cos_theta_list)
    log_angular_Omega = np.log(angular_Omega_tab)
    omega_tab = nu_tab * 2 * pi
    N_freqs = np.size(omega_tab)
    
    def emiss_beta(beta_ind):
        beta = beta_tab[beta_ind]
        aux_ip = mu2_f_ip[beta_ind, :]
        mask_ip = aux_ip > 1e-300
        aux_ip = aux_ip[mask_ip]
        log_angular_Omega_ip = log_angular_Omega[mask_ip]
        interp_ip = interp1d( log_angular_Omega_ip , np.log(aux_ip), kind='cubic', fill_value='extrapolate', bounds_error=False)
        aux_op = mu2_f_op[beta_ind, :]
        mask_op = aux_op > 0
        aux_op = aux_op[mask_op]
        log_angular_Omega_op = log_angular_Omega[mask_op]
        interp_op = interp1d( log_angular_Omega_op , np.log(aux_op), kind='cubic', fill_value='extrapolate', bounds_error=False)
        emiss_mode_1 = 8/9 *  np.exp(interp_op(np.log(omega_tab)))  * omega_tab**4
        if beta == 0: 
            aux = 8/9 *  np.exp(interp_ip(np.log(omega_tab)))  * omega_tab**4
            return aux + emiss_mode_1
        aux = np.zeros((n_cos_theta, N_freqs))
        for theta_ind in range(n_cos_theta):
            cos_theta = cos_theta_list[theta_ind]
            log_omega_tab_mode2 = np.log(omega_tab/np.abs(1+beta*cos_theta))
            log_omega_tab_mode3 = np.log(omega_tab/np.abs(1-beta*cos_theta))
            part2= 1/3 *  np.exp(interp_ip(log_omega_tab_mode2))  * (1+cos_theta)**2 / np.abs(1+beta*cos_theta)
            part3= 1/3 *  np.exp(interp_ip(log_omega_tab_mode3))  * (1-cos_theta)**2 / np.abs(1-beta*cos_theta)
            if cos_theta == 0:
                aux[theta_ind, :] = (part2 + part3) * omega_tab**4
                continue
            log_omega_tab_mode4 = np.log(omega_tab/np.abs(beta*cos_theta))
            part4= 2/3 * np.exp(interp_ip(log_omega_tab_mode4)) * (1-cos_theta**2) / np.abs(beta*cos_theta)
            aux[theta_ind, :] = (part2 + part3 + part4) * omega_tab**4
        return np.average(aux, axis=0, weights=cos_theta_weights[beta_ind]) + emiss_mode_1 # shape (N_freqs,)
    result = parallel_map(emiss_beta, np.arange(len(beta_tab)))
    result = np.array(result)
    result = np.sum(result, axis=0) # shape (N_freqs,)
    return result # shape (N_freqs,)
    